/* This is a helper tool for sysconfig tools as hwup/down, ifup/down and all
 * tools that create device or interface configurations. It may be called with
 * different names for different purposes.
 * For more information see doc/Specification in the sysconfig CVS
 *
 * Version: $Id: get_config.c 1446 2006-03-10 16:14:16Z zoz $
 */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <getopt.h>

#include "get_config.h"
#include "get_config_lib.h"

extern int verbosity;
extern int syslog_verbosity;
extern char syslog_name [MAX_HWDESC_PART_LEN];

char application_name[MAX_MESSAGE_LEN];
extern const char *hwdesc_type_names[];
const char *job_names[] = { JOB_NAMES };

void usage(char *comment) {
    job_t job;
    fprintf(stderr, "Usage: <name> <options> <args>\n"
	   "Call me with one of the following names:\n");
    for (job=0; job<eojl; job++)
	fprintf(stderr, "  %s\n", job_names[job]);
    fprintf(stderr, "Options: (no all of them are usefull for all jobs)\n"
	    "  --config, -c:\n"
	    "      <arg> is a configuration name and not a hardware "
	                                             "description\n" 
	    "  --directory=<dir>, -d <dir>:\n"
	    "      look for configurations in <dir>\n"
	    "      (default: /etc/sysconfig/network\n"
	    "  --file-prefix=<fpfx>, -f <fpfx>:\n"
	    "      use <fpfx> as prefix for configuration files\n"
	    "      (default: 'ifcfg-')\n"
	    "  --help, -h:\n"
	    "      print usage information\n"
	    "  --ifacetype, -i:\n"
	    "      set the interfacetype to look for in getcfg-interface\n"
	    "  --literally, -l:\n"
	    "      take hardware description literally and don't complete it\n"
	    "  --multiarg, -m:\n"
	    "      take multiple arguments like when called as %s\n"
	    "      (see <arg> below)\n"
	    "  --noexport, -e:\n"
	    "      Don't print a final line 'export <list of varaibles>'\n"
	    "  --prefix=<pfx>, p <pfx>:\n"
	    "      Use this prefix for variable names\n"
	    "      (default: 'HWD_')\n"
	    "  --type=<tp>, -t <tp>:\n"
	    "      Choose the type of hardware description for %s\n"
	    "      (default: <none>)\n"
	    "  --verbose=<N>, -v:\n"
	    "      set verbosity to level <N> or use -v multiple times "
	    "(e.g -vvv)\n",
	    job_names[make_hwdesc], job_names[make_hwdesc]);
    fprintf(stderr, "<arg> is exactly one hardware description except when called as\n"
	   "%s. Then there may be any number of arguments all of the\n"
	   "form VARIABLE=ARG. Variable prefixes (HWD_) may be omitted.\n",
	    job_names[make_hwdesc]);
    fprintf(stderr, "\n%s\n", comment);
}

void error_exit(char *file, int line, int retval, const char *comment, ...) {
    char message[3*MAX_MESSAGE_LEN] = {0};
    va_list vl;

    if (retval == 0) {
	sprintf(message, "%s (%s, %d)", application_name,
		file, line);
	perror(message);
	retval = errno;
    }
    else {
	fprintf(stderr,"Error in %s (%s, %d):\n", application_name, file,
		line);
    }
    
    va_start(vl, comment);
    vfprintf(stderr, comment, vl);
    va_end(vl);

    fprintf(stderr, "\n");

    if (retval == 1)
	usage("");

    exit (retval);
}

int main(int argc, char *argv[])
{
    hwdesc_t *hwdesc;
    size_t length = 0;
    char *basename, *property, *copy;
    job_t job = parse_hwdesc;
    int is_confname = 0;
    char confname[MAX_HWDESC_LEN] = {'\0'};
    int export = 1;
    int literally = 0;
    int multiarg = 0;
    char var_prefix[MAX_PREFIX_LEN]; /* = "HWD"; */
    char *directory = NULL;
    char *file_prefix = NULL;
    int retval = 0;
    hwdesc_type_t ht = none;
    hwdesc_interface_class_t interface_type = iface_class_none;
    int busname_num = 0, busid_num = 0;
    int interface_num = 0, ifnum;
    int interface_path_num = 0, interface_type_num = 0;
    char *interface_name;

    char interfacelist[MAX_HWDESC_LEN];
    char interfacetype[MAX_HWDESC_PART_LEN] = "net";

    var_prefix[0] = 0;
    strcpy(var_prefix,"HWD");

    /* At the very first check how we are called:
     */
    basename = rindex(argv[0], '/');
    if (basename == NULL)
	basename = argv[0];
    else
	basename++;
    strncpy(application_name, basename, MAX_MESSAGE_LEN - 1);
    strncpy(syslog_name, basename, MAX_HWDESC_PART_LEN - 1);

    for (job=get_config; job<eojl; job++)
	if (!strcmp(application_name, job_names[job]))
	    break;
    if (job == eojl)
	error_exit(__FILE__, __LINE__, 1, "Don't call me %s!\n",
		   application_name);

    /* We need a struct hwdesc in every case with all set to zero. */
    hwdesc = calloc(1, sizeof(hwdesc_t));
    if (hwdesc == NULL)
	error_exit(__FILE__, __LINE__, 0, "");

    /* Then parse command line for options. We have to do that as second step,
     * because not all options are valid for all jobs.
     */
    while (1) {
	int option_index = 0;
	int optchar, v;
	char *endptr;
	static struct option long_options[] = {
	    {"config", 0, 0,'c'},
	    {"directory", 1, 0, 'd'},
	    {"file-prefix", 1, 0, 'f'},
	    {"help", 0, 0, 'h'},
	    {"ifacetype", 1, 0, 'i'},
	    {"literally", 0, 0, 'l'},
	    {"multiarg", 0, 0, 'm'},
	    {"noexport", 0, 0, 'e'},
	    {"prefix", 1, 0, 'p'},
	    {"type", 1, 0, 't'},
	    {"verbose", 2, 0, 'v'},
	    {0, 0, 0, 0}
	};

	optchar = getopt_long (argc, argv, "cd:ef:hi:lmp:t:v::",
			 long_options, &option_index);
	if (optchar == -1)
	    break;

//	printf("option_index: %d\noption->name: %s\n",
//	       option_index, long_options[option_index].name);

	switch (optchar) {
	    case 0:
		printf ("long only option %s",
			long_options[option_index].name);
		if (optarg)
		    printf (" with arg %s", optarg);
		printf ("\n");
		break;
	    case 'c':
		is_confname = 1;
		break;
	    case 'd':
		directory = optarg;
		break;
	    case 'e':
		export = 0;
		break;
	    case 'f':
		file_prefix = optarg;
		break;
	    case 'h':
		usage("Have fun!");
		return 0;
	    case 'i':
		strncpy(interfacetype, optarg, MAX_HWDESC_PART_LEN - 1);
		break;
	    case 'l':
		literally = 1;
		break;
	    case 'm':
		multiarg = 1;
		break;
	    case 'p':
		strncpy(var_prefix, optarg, MAX_PREFIX_LEN - 1);
		break;
	    case 't':
		hwdesc->hwdesc_type = none;
		for (ht=none; ht<invalid; ht++)
		    if (!strcmp(optarg, hwdesc_type_names[ht]))
			break;
		if (ht == invalid)
		    error_exit(__FILE__, __LINE__, 1,
			       "'%s' is wrong argument for --type\n", optarg);
		hwdesc->hwdesc_type = ht;
		break;
	    case 'v':
		if (optarg) {
		    v = strtol(optarg, &endptr, 0);
		    if ( optarg == endptr ) {
			verbosity++;
			while (*optarg == 'v') {
			    verbosity++;
			    optarg++;
			}
		    }
		    else verbosity+=v;
		}
		else verbosity++;
		break;
	    case '?':
		printf("getopt_long returned '?'\n");
		error_exit(__FILE__, __LINE__, 1, "unknown option\n");
		break;
	    default:
		printf ("?? getopt returned character code 0%o ??\n", optchar);
	}
    }

    if (verbosity > 0)
	printf("%s: verbosity = %d\n", __FUNCTION__, verbosity);
    if (verbosity >= 6) { /* 6 == LOG_INFO */
	printf("%s: is_confname = %d\n", __FUNCTION__, is_confname);
	printf("%s: directory = %s\n", __FUNCTION__, directory);
	printf("%s: export = %d\n", __FUNCTION__, export);
	printf("%s: file_prefix = %s\n", __FUNCTION__, file_prefix);
	printf("%s: literally = %d\n", __FUNCTION__, literally);
	printf("%s: var_prefix = %s\n", __FUNCTION__, var_prefix);
	printf("%s: hwdesc_type = %s\n", __FUNCTION__, hwdesc_type_names[ht]);
    }

/* After we catched all options the remaining arguments will be evaluated.
 * Depending on the job these arguments have different meanings. There are:
 * - hardware description (at most one)
 * - configuration name (at most one)
 * - hardware properties in the form <prefix>_<property name>=<property> (many)
 * - interface name (is that just a hwdesc???)
 * Therefore we check each argument if it starts with '<prefix>_' and contains
 * a '='. If yes, we feed struct hwdesc with that information. If not it
 * depends on the current job and given options:
 * -' parse_hwdesc' takes exactly one hwdesc
 * -' make_hwdesc' takes only hardware properties
 * - 'get device' and
 * - 'get_interface' take a hwdesc or a confname depending on <is_confname>
 * - 'get_all_info' and
 * - 'get_config' again take exactly one hwdesc
*/
    
    switch (job) {
	case get_config:
	case parse_hwdesc:
	case get_all_info:
	    is_confname = 0;
	case get_device:
	case get_interface:
	    break;
	case make_hwdesc:
	    multiarg = 1;
	    break;
	default:
	    error_exit(__FILE__, __LINE__, 77, "'job' out of range\n");
    }

    if (multiarg == 0) {
	if (optind + 1 == argc)
	    if (is_confname)
		strncpy(confname, argv[optind], MAX_HWDESC_LEN -1);
	    else {
		length = strlen(argv[optind]);
		if (length > MAX_HWDESC_LEN)
		    error_exit(__FILE__, __LINE__, 1,
			       "hwdesc to long (max. MAX_HWDESC_LEN "
			       "characters)\n");
		strncpy(hwdesc->hwdesc, argv[optind], MAX_HWDESC_LEN);
	    }
	else
	    error_exit(__FILE__, __LINE__, 1,
		       "Give exactly one argument\n");
    }
    else {
	/* Try to split each argument in prefix, property name and
	 * property: <prefix>_<property name>=<property>. If that succeeds
	 * and the property name is valid, then take the property value.
	 * Don't care about the prefix. It it fails, exit.
	 */
	while (optind < argc) {
	    /* We assume that the property name does not contain a '=' nor
	     * a '_'. Therefore all after the '=' is the property and all
	     * before and including the '_' is the optional prefix. 
	     */
	    property = index(argv[optind], '=');
	    if (property == NULL)
		error_exit(__FILE__, __LINE__, 1, "argument is no valid "
			   "hardware property (no '=' in the string)\n");
	    *property = '\0';
	    property++;
	    basename = index(argv[optind], '_');
	    if (basename == NULL)
		basename = argv[optind];
	    else {
		*basename = '\0';
		basename++;
	    }
	    copy = property;
	    /* Now compare basename with all valid property names */
	    if (!strcmp(basename, "DEVTYPE")) {
		copy = strncpy(hwdesc->devtype, property,
			       MAX_HWDESC_PART_LEN - 1);
		if (*hwdesc->devtype == '\0')
		    *hwdesc->devtype = EMPTY_STRING;
	    }
	    else if (!strcmp(basename, "INTERFACE")) {
		if (interface_num >= MAX_INTERFACE_NUM)
		    error_exit(__FILE__, __LINE__, 1, "too many interfaces.\n");
		copy = strncpy(hwdesc->interface[interface_num],
			       property, MAX_HWDESC_PART_LEN - 1);
		interface_num++;
	    }
	    else if (!strcmp(basename, "ID")) 
		copy = strncpy(hwdesc->id, property,
			       MAX_HWDESC_PART_LEN - 1);
	    /* number of busnames and busids have to be checked at the
	     * very end, if #busid <= #busname <= #busid + 1 */
	    else if (!strcmp(basename, "BUSNAME")) {
		if (busname_num >= MAX_INTERFACE_NUM)
		    error_exit(__FILE__, __LINE__, 1, "to many busnames.\n");
		copy = strncpy(hwdesc->busname[busname_num], property,
			       MAX_HWDESC_PART_LEN - 1);
		busname_num++;
	    }
	    else if (!strcmp(basename, "BUSID")) {
		if (busid_num >= MAX_INTERFACE_NUM)
		    error_exit(__FILE__, __LINE__, 1, "to much busids.\n");
		copy = strncpy(hwdesc->busid[busid_num],
			       property, MAX_HWDESC_PART_LEN - 1);
		busid_num++;
	    }
	    else if (!strcmp(basename, "VENDORID")) {
		copy = strncpy(hwdesc->vendorid, property,
			       MAX_HWDESC_PART_LEN - 1);
		if (*hwdesc->vendorid == '\0')
		    *hwdesc->vendorid = EMPTY_STRING;
	    }
	    else if (!strcmp(basename, "VENDORSUBID")) {
		copy = strncpy(hwdesc->vendorsubid, property,
			       MAX_HWDESC_PART_LEN - 1);
		if (*hwdesc->vendorsubid == '\0')
		    *hwdesc->vendorsubid = EMPTY_STRING;
	    }
	    else if (!strcmp(basename, "PRODUCTID")) {
		copy = strncpy(hwdesc->productid, property,
			       MAX_HWDESC_PART_LEN - 1);
		if (*hwdesc->productid == '\0')
		    *hwdesc->productid = EMPTY_STRING;
	    }
	    else if (!strcmp(basename, "PRODUCTSUBID")) {
		copy = strncpy(hwdesc->productsubid, property,
			       MAX_HWDESC_PART_LEN - 1);
		if (*hwdesc->productsubid == '\0')
		    *hwdesc->productsubid = EMPTY_STRING;
	    }
	    else if (!strcmp(basename, "VPIDREV")) {
		copy = strncpy(hwdesc->vpidrev, property,
			       MAX_HWDESC_PART_LEN - 1);
		if (*hwdesc->vpidrev == '\0')
		    *hwdesc->vpidrev = EMPTY_STRING;
	    }
	    else if (!strcmp(basename, "DEVPATH")) 
		copy = strncpy(hwdesc->devpath, property,
			       MAX_DEVPATH_LEN - 1);
	    else if (!strcmp(basename, "DEVICEPATH")) 
		copy = strncpy(hwdesc->devicepath, property,
			       MAX_DEVPATH_LEN - 1);
	    else if (!strcmp(basename, "INTERFACEPATH")) {
		if (interface_path_num >= MAX_INTERFACE_NUM)
		    error_exit(__FILE__, __LINE__, 1, "to many interface paths.\n");
		copy = strncpy(hwdesc->interfacepath[interface_path_num],
			       property, MAX_HWDESC_PART_LEN - 1);
		interface_path_num++;
	    } else if (!strcmp(basename, "INTERFACETYPE")) {
		if (interface_type_num >= MAX_INTERFACE_NUM)
		    error_exit(__FILE__, __LINE__, 1, "to many interface types.\n");
		copy = strncpy(hwdesc->interfacepath[interface_type_num],
			       property, MAX_HWDESC_PART_LEN - 1);
		interface_type_num++;
	    } else
		error_exit(__FILE__, __LINE__, 1,
			   "unknown hardware property name '%s'\n",
			   basename);
	    if (strcmp(property, copy) && *copy != EMPTY_STRING)
		error_exit(__FILE__, __LINE__, 1,
			   "value for hardware property name '%s' to "
			   "long:\n%s\n", basename, property);
	    optind++;
	}
	/* Check number of busnames and busids, if 
	 * #busid <= #busname <= #busid + 1
	 * Set hwdesc->bus_num appropriately. */
	if (busid_num > busname_num)
	    error_exit(__FILE__, __LINE__, 1, "Too many busids\nThere "
		       "must be no more busids (%d) then busnames (%d)\n",
		       busid_num, busname_num);
	if (busid_num < busname_num -1)
	    error_exit(__FILE__, __LINE__, 1, "Too many busnames\nThere must be"
	               "at most one busnames (%d) more than busids (%d)\n",
	    busname_num, busid_num);

	hwdesc->bus_num = busname_num;

	/* Sanitize any given interfaces */
	if (interface_type_num > interface_num)
	    error_exit(__FILE__, __LINE__, 1, "Too many interface types\n"
		       "There must be no more interface types (%d) "
		       "than interface names (%d)\n",
		       interface_type_num, interface_num);
	for ( ifnum = 0; ifnum < interface_num; ifnum++ ) {
	    interface_type = match_interface_type(hwdesc->interface[ifnum]);
	    interface_name = get_interface_name(interface_type);
	    if (hwdesc->interfaceclass[ifnum] != iface_class_none &&
		hwdesc->interfaceclass[ifnum] != interface_type) {
		printf("Interface %d (%s) has wrong type '%s', "
		       "correcting to '%s'\n", ifnum, hwdesc->interface[ifnum],
		       get_interface_name(hwdesc->interfaceclass[ifnum]),
		       interface_name);
		hwdesc->interfaceclass[ifnum] = interface_type;
	    }
	    if (*hwdesc->interfacetype[ifnum] &&
		strcmp(hwdesc->interfacetype[ifnum], interface_name)) {
		printf("Interface %d (%s) has wrong type '%s', "
		       "correcting to '%s'\n", ifnum, hwdesc->interface[ifnum],
		       hwdesc->interfacetype[ifnum], interface_name);
		strcpy(hwdesc->interfacetype[ifnum], interface_name);
	    }
	}
    }
	
    /* Now finally execute the wanted job */
    switch (job) {
	case parse_hwdesc:
	    if (multiarg == 0)
		retval = split_hwdesc(hwdesc);
	    break;
	case get_all_info:
	    if (multiarg == 0)
		retval = split_hwdesc(hwdesc);
	    if (retval) break;
	    retval = complete_hwdesc_sysfs(hwdesc);
	    break;
	case get_config:
	    if (multiarg == 0)
		retval = split_hwdesc(hwdesc);
	    if (retval) break;
	    if (literally == 0)
		retval = complete_hwdesc_sysfs(hwdesc);
	    if (retval) break;
	    retval = get_configuration(hwdesc, directory, file_prefix);
	    break;
	case get_device:
	    printf("get_device not yet implemented\n");
	    retval = 1;
	    break;
	case get_interface:
	    if (multiarg == 0)
		retval = split_hwdesc(hwdesc);
//	    if (retval) break;
	    if (literally == 0)
		retval = complete_hwdesc_sysfs(hwdesc);
//	    if (retval) break;
//	    printf("interface: %s\n", hwdesc->interface[0]);

	    look_smpppd_info(hwdesc->interface[0]);

	    if (*hwdesc->interface[0] != '\0' &&
		0 <= match_type(hwdesc->devtype, interfacetype)) {
		printf("%s\n", hwdesc->interface[0]);
		return 0;
	    }
//	    get_all_interfaces("net", "", interfacelist);
//	    printf("interfacelist: %s\n", interfacelist);
	    if (1 == get_all_interfaces(interfacetype, hwdesc->hwdesc,
				       interfacelist)) {
		if (verbosity == 0)
		    printf("%s\n", interfacelist);
		else
		    printf("%s (indirekt)\n", interfacelist);
		return 0;
	    }
	    return 11;
	case make_hwdesc:
	    if (hwdesc->hwdesc_type == none)
		error_exit(__FILE__, __LINE__, 1, "Use option '--type' to "
			   "choose a hardware description type.\n");
	    retval = complete_hwdesc_sysfs(hwdesc);
	    if (retval) break;
	    retval = assemble_hwdesc(hwdesc);
	case eojl:  /* These two lines just avoid ... */
	    break;  /* ... a compiler warning. */
    }

    print_hwdesc(hwdesc, var_prefix, export);

    return retval;
}

/*

complete_hwdesc_sysfs(hwdesc);
401	printf("Cannot get sysfs mnt path");

11: getcfg-interface did not find an interface

*/
