/* This is the library behind the helper tool for sysconfig tools. It provides
 * functions that parse or assemble hardware descriptions or configuration
 * names and get as much as possible information from sysfs.
 * For more information see doc/Specification in the sysconfig CVS
 *
 * Version: $Id: get_config_lib.c 1458 2006-03-17 11:12:30Z zoz $
 */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <getopt.h>
#include <dirent.h>
#include <syslog.h>
#include <ctype.h>

#include "get_config_lib.h"

// #define	LOG_EMERG	0	/* system is unusable */
// #define	LOG_ALERT	1	/* action must be taken immediately */
// #define	LOG_CRIT	2	/* critical conditions */
// #define	LOG_ERR		3	/* error conditions */
// #define	LOG_WARNING	4	/* warning conditions */
// #define	LOG_NOTICE	5	/* normal but significant condition */
// #define	LOG_INFO	6	/* informational */
// #define	LOG_DEBUG	7	/* debug-level messages */

int verbosity = LOG_EMERG;
int syslog_verbosity = LOG_EMERG;
char syslog_name [MAX_HWDESC_PART_LEN] = {'\0'};
const char *hwdesc_type_names[] = { HWDESC_TYPE_NAMES };
const char *hwdesc_bus_names[] = { HWDESC_BUS_NAMES };

void error_log(int level, const char* message, ...) {
    va_list vl;
    if (level <= verbosity) {
	va_start(vl, message);
	vfprintf(stderr, message, vl);
	va_end(vl);
    }
    if (level <= syslog_verbosity) {
	va_start(vl, message);
	openlog(syslog_name, LOG_CONS|LOG_PID, LOG_USER);
	syslog(level, "%s", message);
	closelog();
	va_end(vl);
    }
}

/*
 * Logging interface for libsysfs
 *
 * If libsysfs is compiled with -DLOG, it expects a function
 * log_message() from the caller.
 */
void log_message(int level, const char *format, ...)
{
        va_list args;

        va_start(args, format);
        error_log(level, format, args);
        va_end(args);
}

void print_hwdesc(hwdesc_t *hwdesc, char *prefix, int export) {
    int i;
    char varlist[MAX_INTERFACE_NUM*80] = {'\0'};
    char varname[MAX_INTERFACE_NUM*80] = {'\0'};

    /* It may possibly harm to concatenate strings with sprintf(). But while
     * we always just append something, it will probably work.
     */
    if (hwdesc->hwdesc[0] != '\0') {
	printf("%s_HWDESC=\"%s\";\n", prefix, hwdesc->hwdesc);
	sprintf(varname, " %s_HWDESC", prefix);
	strcat(varlist, varname);
    }
    if (hwdesc->hwdesc_type != none) {
	printf("%s_HWDESCTYPE=\"%s\";\n", prefix,
	       hwdesc_type_names[hwdesc->hwdesc_type]);
	sprintf(varname, " %s_HWDESCTYPE", prefix);
	strcat(varlist, varname);
    }
    if (hwdesc->devtype[0] != '\0') {
	printf("%s_DEVTYPE=\"%s\";\n", prefix, hwdesc->devtype);
	sprintf(varname, " %s_DEVTYPE", prefix);
	strcat(varlist, varname);
    }
//    if (hwdesc->interface_num) {
	printf("%s_INTERFACE_N=\"%d\";\n", prefix, hwdesc->interface_num);
	sprintf(varname, " %s_INTERFACE_N", prefix);
	strcat(varlist, varname);
//    }
    for (i=0; i<MAX_INTERFACE_NUM; i++)
	if (hwdesc->interface[i][0] != '\0') {
	    printf("%s_INTERFACE_%d=\"%s\";\n", prefix, i, hwdesc->interface[i]);
	    sprintf(varname, " %s_INTERFACE_%d", prefix, i);
	    strcat(varlist, varname);
	    printf("%s_INTERFACEPATH_%d=\"%s\";\n", prefix, i, 
		   hwdesc->interfacepath[i]);
	    sprintf(varname, " %s_INTERFACEPATH_%d", prefix, i);
	    strcat(varlist, varname);
	    printf("%s_INTERFACETYPE_%d=\"%s\";\n", prefix, i, 
		   hwdesc->interfacetype[i]);
	    sprintf(varname, " %s_INTERFACETYPE_%d", prefix, i);
	    strcat(varlist, varname);
	}
    if (hwdesc->id[0] != '\0') {
	printf("%s_ID=\"%s\";\n", prefix, hwdesc->id);
	sprintf(varname, " %s_ID", prefix);
	strcat(varlist, varname);
    }
//    if (hwdesc->bus_num) {
	printf("%s_BUS_N=\"%d\";\n", prefix, hwdesc->bus_num);
	sprintf(varname, " %s_BUS_N", prefix);
	strcat(varlist, varname);
//    }
    for (i=0; i<hwdesc->bus_num; i++) {
	    printf("%s_BUSNAME_%d=\"%s\";\n", prefix, i, hwdesc->busname[i]);
	    sprintf(varname, " %s_BUSNAME_%d", prefix, i);
	    strcat(varlist, varname);
	    printf("%s_BUSID_%d=\"%s\";\n", prefix, i, hwdesc->busid[i]);
	    sprintf(varname, " %s_BUSID_%d", prefix, i);
	    strcat(varlist, varname);
    }
    if (hwdesc->vendorid[0] != '\0') {
	printf("%s_VENDORID=\"%s\";\n", prefix, hwdesc->vendorid);
	sprintf(varname, " %s_VENDORID", prefix);
	strcat(varlist, varname);
    }
    if (hwdesc->productid[0] != '\0') {
	printf("%s_PRODUCTID=\"%s\";\n", prefix, hwdesc->productid);
	sprintf(varname, " %s_PRODUCTID", prefix);
	strcat(varlist, varname);
    }
    if (hwdesc->vendorsubid[0] != '\0') {
	printf("%s_VENDORSUBID=\"%s\";\n", prefix, hwdesc->vendorsubid);
	sprintf(varname, " %s_VENDORSUBID", prefix);
	strcat(varlist, varname);
    }
    if (hwdesc->productsubid[0] != '\0') {
	printf("%s_PRODUCTSUBID=\"%s\";\n", prefix, hwdesc->productsubid);
	sprintf(varname, " %s_PRODUCTSUBID", prefix);
	strcat(varlist, varname);
    }
    if (hwdesc->vpidrev[0] != '\0') {
	printf("%s_VPIDREV=\"%s\";\n", prefix, hwdesc->vpidrev);
	sprintf(varname, " %s_VPIDREV", prefix);
	strcat(varlist, varname);
    }
    if (hwdesc->driver[0] != '\0') {
	printf("%s_DRIVER=\"%s\";\n", prefix, hwdesc->driver);
	sprintf(varname, " %s_DRIVER", prefix);
	strcat(varlist, varname);
    }
//    if (hwdesc->config_num) {
	printf("%s_CONFIG_N=\"%d\";\n", prefix, hwdesc->config_num);
	sprintf(varname, " %s_CONFIG_N", prefix);
	strcat(varlist, varname);
//    }
    for (i=0; i<MAX_INTERFACE_NUM; i++)
	if (hwdesc->config[i][0] != '\0') {
	    printf("%s_CONFIG_%d=\"%s\";\n", prefix, i, hwdesc->config[i]);
	    sprintf(varname, " %s_CONFIG_%d", prefix, i);
	    strcat(varlist, varname);
	}
    if (hwdesc->devpath[0] != '\0') {
	printf("%s_DEVPATH=\"%s\";\n", prefix, hwdesc->devpath);
	sprintf(varname, " %s_DEVPATH", prefix);
	strcat(varlist, varname);
    }
    if (hwdesc->devicepath[0] != '\0') {
	printf("%s_DEVICEPATH=\"%s\";\n", prefix, hwdesc->devicepath);
	sprintf(varname, " %s_DEVICEPATH", prefix);
	strcat(varlist, varname);
    }
    if (export)
	printf("export %s;\n", varlist);
}

/* reset whole struct except the hardware description itself */
void reset_hwdesc(hwdesc_t *hwdesc) {
    memset(hwdesc->devtype, 0, MAX_HWDESC_PART_LEN);
    memset(hwdesc->interface, 0, MAX_INTERFACE_NUM * MAX_HWDESC_PART_LEN);
    hwdesc->interface_num = 0;
    memset(hwdesc->id, 0, MAX_HWDESC_PART_LEN);
    memset(hwdesc->busname, 0, MAX_INTERFACE_NUM * MAX_HWDESC_PART_LEN);
    memset(hwdesc->busid, 0, MAX_INTERFACE_NUM * MAX_HWDESC_PART_LEN);
    hwdesc->bus_num = 0;
    memset(hwdesc->vendorid, 0, MAX_HWDESC_PART_LEN);
    memset(hwdesc->productid, 0, MAX_HWDESC_PART_LEN);
    memset(hwdesc->vendorsubid, 0, MAX_HWDESC_PART_LEN);
    memset(hwdesc->productsubid, 0, MAX_HWDESC_PART_LEN);
    memset(hwdesc->vpidrev, 0, MAX_HWDESC_PART_LEN);
    memset(hwdesc->config, 0, MAX_INTERFACE_NUM * MAX_HWDESC_PART_LEN);
    hwdesc->config_num = 0;
    hwdesc->hwdesc_type = none;
    memset(hwdesc->devpath, 0, MAX_DEVPATH_LEN);
    memset(hwdesc->devicepath, 0, MAX_DEVPATH_LEN);
    memset(hwdesc->interfacepath, 0, MAX_INTERFACE_NUM * MAX_DEVPATH_LEN);
    memset(hwdesc->interfacetype, 0, MAX_INTERFACE_NUM * MAX_DEVPATH_LEN);

    return;
}

int parse_vpid_hwargs(hwdesc_t *hwdesc, char *hwarg)
{
    int args_parsed = 0;
    char *vpid_part, *vpid_str;
    size_t vpid_part_len;


    /* split off the argument */
    vpid_part = hwarg;
    error_log(LOG_DEBUG,"vpid_part '%s'\n",vpid_part);
    vpid_str = strchr(vpid_part, '-');
    if (!vpid_str) {
	/* Just one argument */
	strcpy(hwdesc->vendorid, vpid_part);
	return args_parsed;
    }
    vpid_part_len = (int)(vpid_str - vpid_part);
    strncpy(hwdesc->vendorid, vpid_part, vpid_part_len);
    hwdesc->vendorid[vpid_part_len] = '\0';
    args_parsed++;

    vpid_part = vpid_str + 1;
    error_log(LOG_DEBUG,"vpid_part '%s'\n",vpid_part);
    vpid_str = strchr(vpid_part, '-');
    if (!vpid_str) {
	strcpy(hwdesc->productid, vpid_part);
	return args_parsed;
    }
    vpid_part_len = (int)(vpid_str - vpid_part);
    strncpy(hwdesc->productid, vpid_part, vpid_part_len);
    hwdesc->productid[vpid_part_len] = '\0';
    args_parsed++;

    vpid_part = vpid_str + 1;
    error_log(LOG_DEBUG,"vpid_part '%s'\n",vpid_part);
    vpid_str = strchr(vpid_part, '-');
    if (!vpid_str) {
	strcpy(hwdesc->vendorsubid, vpid_part);
	return args_parsed;
    }
    vpid_part_len = (int)(vpid_str - vpid_part);
    strncpy(hwdesc->vendorsubid, vpid_part, vpid_part_len);
    hwdesc->vendorsubid[vpid_part_len] = '\0';
    args_parsed++;

    vpid_part = vpid_str + 1;
    error_log(LOG_DEBUG,"vpid_part '%s'\n",vpid_part);
    vpid_str = strchr(vpid_part, '-');
    if (!vpid_str) {
	strcpy(hwdesc->productsubid, vpid_part);
	return args_parsed;
    }
    vpid_part_len = (int)(vpid_str - vpid_part);
    strncpy(hwdesc->productsubid, vpid_part, vpid_part_len);
    hwdesc->productsubid[vpid_part_len] = '\0';
    args_parsed++;

    vpid_part = vpid_str + 1;
    error_log(LOG_DEBUG,"vpid_part '%s'\n",vpid_part);
    strcpy(hwdesc->vpidrev, vpid_part);

    return args_parsed;
}

/*
 * get_bus_type
 *
 * Returns the type number of the bus name at the beginning
 * of the string 'hwarg'.
 */
hwdesc_bus_t get_bus_type(char *hwarg)
{
    hwdesc_bus_t bus_type;

    for (bus_type = bus_none; bus_type < bus_invalid; bus_type++) {
	if (!strncmp(hwarg, hwdesc_bus_names[bus_type],
		     strlen(hwdesc_bus_names[bus_type]))) {
	    break;
	}
    }

    return bus_type;
}

/*
 * Search for the next valid bus name.
 * A valid bus name is a known bus name preceded by a '-'.
 * Returns a pointer to the separator preceding the found bus
 * name and sets 'ret_bus_type' accordingly.
 * If no bus name has been found it returns NULL and sets
 * 'ret_bus_type' to 'bus_invalid'.
 */
char *search_next_bus_type(char *hwarg, hwdesc_bus_t *ret_bus_type)
{
    hwdesc_bus_t bus_type;
    char hwsearch[MAX_HWDESC_LEN] = {'\0'};
    char *bus_ptr = NULL;

    *ret_bus_type = bus_invalid;
    for (bus_type = bus_none; bus_type < bus_invalid; bus_type++) {
	strcpy(hwsearch, "-");
	strcat(hwsearch, hwdesc_bus_names[bus_type]);
	if ((bus_ptr = strstr(hwarg, hwsearch))) {
	    *ret_bus_type = bus_type;
	    break;
	}
    }
    return bus_ptr;
}

/*
 * recursively parse hwargs for bus description.
 * Expects 'hwarg' to start with the bus name
 * designated by 'bus_type'.
 */
int parse_bus_hwargs_recursive(hwdesc_t *hwdesc, char *hwarg,
			       hwdesc_bus_t bus_type)
{
    char *bus_part, *bus_ptr;
    int retval;

    error_log(LOG_DEBUG,"%s: num %d arg '%s' type %d\n",__FUNCTION__,
	      hwdesc->bus_num, hwarg, bus_type);

    /* We have found a valid bus name */
    strcpy(hwdesc->busname[hwdesc->bus_num], hwdesc_bus_names[bus_type]);
    bus_part = hwarg + strlen(hwdesc_bus_names[bus_type]);

    error_log(LOG_DEBUG,"bus_part '%s'\n",bus_part);

    if (bus_part[0] == '\0') {
	/* No bus id, okay */
	hwdesc->bus_num++;
	goto out;
    }

    if (bus_part[0] != '-') {
	/* bus id does not start with '-', error */
	hwdesc->hwdesc_type = invalid;
	hwdesc->busname[hwdesc->bus_num][0] = '\0';
    } else {
	bus_ptr = search_next_bus_type(bus_part, &bus_type);
	/* skip the separator _after_ we've search for the
	 * next bus type. This enables us to catch
	 * bus-pci-ide etc. */
	bus_part++;

	error_log(LOG_DEBUG,"bus_part '%s' bus_ptr '%s' type %d\n",
		  bus_part, bus_ptr, bus_type);

	if (bus_ptr) {
	    /* Valid bus type found */
	    strncpy(hwdesc->busid[hwdesc->bus_num], bus_part,
		    (int)(bus_ptr - bus_part));
	    hwdesc->busid[hwdesc->bus_num][bus_ptr - bus_part] = '\0';
	    hwdesc->bus_num++;
	    retval = parse_bus_hwargs_recursive(hwdesc, bus_ptr + 1, bus_type);
	} else {
	    /* No valid bus type found, argument is the bus id */
	    if (bus_part[strlen(bus_part) - 1] == '-') {
		/* Delete trailing separator */
		bus_part[strlen(bus_part) - 1] = '\0';
	    }
	    strcpy(hwdesc->busid[hwdesc->bus_num], bus_part);
	    hwdesc->bus_num++;
	}
    }
    
 out:
    return hwdesc->bus_num;
}

/*
 * parse_bus_hwargs
 *
 * Parse the bus identification string 'hwarg' into
 * the hardware description 'hwdesc'.
 * Just sets up the string and calls the recursive
 * parser.
 */
int parse_bus_hwargs(hwdesc_t *hwdesc, char *hwarg)
{
    int retval = 0;
    hwdesc_bus_t bus_type;

    error_log(LOG_DEBUG,"%s: arg '%s'\n",__FUNCTION__, hwarg);

    bus_type = get_bus_type(hwarg);
    if (bus_type == bus_invalid || bus_type == bus_none) {
	/* Invalid or no bus name */
	hwdesc->hwdesc_type = invalid;
	return -1;
    }

    retval = parse_bus_hwargs_recursive(hwdesc, hwarg, bus_type);

    return retval;
}

int split_hwdesc(hwdesc_t *hwdesc)
{
    char hsearch[MAX_HWDESC_LEN] = {'\0'};
    char devtype[MAX_HWDESC_LEN] = {'\0'};
    char *hwarg = NULL, *hwdesc_ptr;
    size_t hwdesc_length = strlen(hwdesc->hwdesc), devtype_len;
    hwdesc_type_t htype;

    reset_hwdesc(hwdesc);

    error_log(LOG_DEBUG,"%s\n",__FUNCTION__);

    /*
     * First we look for known type identifiers:
     * 'id-'
     * 'bus-'
     * 'devpath-'
     * 'interface-'
     * 'vpid-'
     * 'type-'
     */

    /*
     * First step:
     * type identifier at the start of hwdesc
     */
    for (htype=none; htype<invalid; htype++) {
	strcpy(hsearch, hwdesc_type_names[htype]);
	strcat(hsearch, "-");
	if ((!strncmp(hwdesc->hwdesc, hsearch, strlen(hsearch)))) {
	    /* identifier at the start of hwdesc, no devtype */
	    devtype_len = 0;
	    devtype[0] = '\0';
	    hwarg = hwdesc->hwdesc + strlen(hsearch);
	    break;
	}
    }

    error_log(LOG_DEBUG, "found type %d\n", htype);

    if (htype == invalid) {
	/*
	 * Second step:
	 * Type identifier within the string
	 */
	for (htype=none; htype<invalid; htype++) {
	    strcpy(hsearch, "-");
	    strcat(hsearch, hwdesc_type_names[htype]);
	    strcat(hsearch, "-");
	    if ((hwdesc_ptr = strstr(hwdesc->hwdesc, hsearch))) {
		devtype_len = (int)(hwdesc_ptr - hwdesc->hwdesc);
		/* Strip off any 'static-' at the start */
		if (devtype_len && !strncmp(hwdesc->hwdesc,"static-",7)) {
		    devtype_len -= 6;
		    strncpy(devtype,hwdesc->hwdesc + 7, devtype_len);
		} else {
		    strncpy(devtype,hwdesc->hwdesc,devtype_len);
		}
		devtype[devtype_len] = '\0';
		if (strlen(hwdesc_ptr) > strlen(hsearch))
		    hwarg = hwdesc_ptr + strlen(hsearch);
		else
		    hwarg = NULL;
		break;
	    }
	}
    }

    error_log(LOG_DEBUG,"'%s' - type %d - '%s'\n",
	      devtype, htype, hwarg);

    if (!hwarg || hwarg[0] == '\0') {
	/* type identifier needs to be followed by a description */
	hwdesc->hwdesc_type = invalid;
    } else {
	hwdesc->hwdesc_type = htype;
    }

    /* parse the description */
    switch (hwdesc->hwdesc_type) {
	case none:
	    hwdesc->hwdesc_type = invalid;
	    break;
	case id:
	    /* The entire description is valid, no specific parsing */
	    strcpy(hwdesc->id, hwarg);
	    break;
	case vpid:
	    /* split off the argument */
	    parse_vpid_hwargs(hwdesc, hwarg);
	    break;
	case bus:
	    parse_bus_hwargs(hwdesc, hwarg);
	    break;
	case devpath:
	    /* The entire description is valid, no specific parsing */
	    strcpy(hwdesc->devpath, hwarg);
	    break;
	case type:
	    /* The entire description is valid, no specific parsing */
	    strcpy(hwdesc->devtype, hwarg);
	    break;
	case interface:
	    if (hwdesc->interface_num < MAX_INTERFACE_NUM) {
		strcpy(hwdesc->interface[hwdesc->interface_num++], hwarg);
		hwdesc->hwdesc_type = interface;
	    } else {
		hwdesc->hwdesc_type = invalid;
	    }
	case invalid:
	    /*
	     * No keywords found.
	     * Assume it's an interface, except if it's empty.
	     */
	    if (    hwdesc->interface_num < MAX_INTERFACE_NUM
	         && hwdesc->hwdesc[0] != '\0'                 ) {
		strcpy(hwdesc->interface[hwdesc->interface_num++],
		       hwdesc->hwdesc);
		hwdesc->hwdesc_type = interface;
	    } else {
		hwdesc->hwdesc_type = invalid;
	    }
	    break;
    }

    /* Store an optional devtype */
    if (hwdesc->hwdesc_type != invalid && devtype[0]) {
	strcpy(hwdesc->devtype, devtype);
    }

    return 0;
}

/* add_string_if_exists()
 * adds the string 'part' with a seperator '-' to 'target'. If
 * 'part' is empty, no seperator is added.
 * 'target' must provide space for MAX_HWDESC_LEN characters.
 * It returns:
 * 0 if 'part' was quoted and added successfully
 * 1 if 'part' is empty
 * -512 if 'target' would exceed MAX_HWDESC_LEN characters. In this case
 * 'target' is not modified.
 */
int add_string_if_exists(char *target, char* part)
{
    int len;

    if (!part || *part == '\0')
	return 1;
    len = strlen(target);
    if ((len + strlen(part) + 1) >= MAX_HWDESC_LEN)
	return -512;

    strcat(target,"-");
    strcat(target,part);

    return 0;
}

/* assemble_hwdesc() creates a new hardware description out of the information
 * inside the given struct hwdesc. In case of success it writes the new
 * description as well into the struct. The type of the new description is also
 * taken from the struct.
 * Except for types 'interface' and 'type', the 'devtype' will be added as
 * first part, if it is available. No error, if it is missing.
 * It returns:
 * 0 if success
 * -1 if there was no valid type for the new description
 * >0 if some needed data was missing
 * -512 if the new description would exceed MAX_HWDESC_LEN characters
 * -511, -510, ... like -512 and some data missing
 */
int assemble_hwdesc(hwdesc_t *hwdesc) {
    char new_hwdesc[MAX_HWDESC_LEN] = {'\0'};
    int i, r = 0, R = 0;

    error_log(LOG_DEBUG,"%s\n",__FUNCTION__);

    switch (hwdesc->hwdesc_type) {
	case id:
	    add_string_if_exists(new_hwdesc, hwdesc->devtype);
	    r += add_string_if_exists(new_hwdesc, "id");
	    if (r) return r;
	    r += add_string_if_exists(new_hwdesc, hwdesc->id);
	    if (r) return r;
	    break;
	case vpid:
	    add_string_if_exists(new_hwdesc, hwdesc->devtype);
	    r += add_string_if_exists(new_hwdesc, "vpid");
//	    r += add_string_if_exists(new_hwdesc, hwdesc->vpid);
//	    if (r) return r;
	    r += add_string_if_exists(new_hwdesc, hwdesc->vendorid);
	    if (r) return r;
	    /* productid, vendorsubid and productsubid are optional. But if any
	     * of these is added, all previous parts have to be added as well*/
	    r = add_string_if_exists(new_hwdesc, hwdesc->productid);
	    if (r < 0) return r;
	    R = r;
	    r = add_string_if_exists(new_hwdesc, hwdesc->vendorsubid);
	    if (r < 0) return r;
	    if (r == 0 && R != 0) return 1;
	    R = r;
	    r = add_string_if_exists(new_hwdesc, hwdesc->productsubid);
	    if (r < 0) return r;
	    if (r == 0 && R != 0) return 1;
	    R = r;
	    r = add_string_if_exists(new_hwdesc, hwdesc->vpidrev);
	    if (r < 0) return r;
	    if (r == 0 && R != 0) return 1;
	    break;
	case bus:
	    add_string_if_exists(new_hwdesc, hwdesc->devtype);
	    r += add_string_if_exists(new_hwdesc, "bus");
	    r += add_string_if_exists(new_hwdesc, hwdesc->busname[0]);
	    if (r) return r;
	    /* busids and additional busnames are optional. But if any
	     * of these is added, all previous parts have to be added as well*/
	    r = add_string_if_exists(new_hwdesc, hwdesc->busid[0]);
	    if (r < 0) return r;
	    R = r;
	    for (i=1; i < hwdesc->bus_num; i++) {
		r = add_string_if_exists(new_hwdesc, hwdesc->busname[i]);
		if (r < 0) return r;
		if (r == 0 && R != 0) return 1;
		R = r;
		r = add_string_if_exists(new_hwdesc, hwdesc->busid[i]);
		if (r < 0) return r;
		if (r == 0 && R != 0) return 1;
		R = r;
	    }
	    break;
	case interface:
	    r += add_string_if_exists(new_hwdesc, hwdesc->interface[0]);
	    if (r) return r;
	    break;
	case type:
	    r += add_string_if_exists(new_hwdesc, "type");
	    r += add_string_if_exists(new_hwdesc, hwdesc->devtype);
	    if (r) return r;
	    break;
	case devpath:
	    add_string_if_exists(new_hwdesc, hwdesc->devtype);
	    r += add_string_if_exists(new_hwdesc, "devpath");
	    r += add_string_if_exists(new_hwdesc, hwdesc->devpath);
	    if (r) return r;
	    break;
	case invalid:
	case none:
	    return -1;
    }
    strcpy(hwdesc->hwdesc, new_hwdesc);
    return 0;
}

/* Compares 'cfg' with 'hwd' and returns the quality of the match:
 * 'cfg' must be completely in 'hwd', 'hwd' must start with 'cfg'. The quality
 * of the match is the length of 'cfg'.
 * If 'cfg' starts with EMPTY_STRING it always matches, but with quality 0.
 *
 * If 'cis' != 0 then it compares case insensitive, which means:
 * If cfg[i] > hwd[i] then subtract a-A from it and compare again
 * If cfg[i] < hwd[i] then add a-A from it and compare again
 * Therfore 0-9 will match P-Y and so on, but i assume we don't need it
 * more exactly. We will see ;)
 *
 * The return value is the quality of the match, or -1 if they don't match.
 */
int match_id(const char *cfg, const char *hwd, int cis, int hex) {
    int case_shift = (cis == 0) ? 0 : 'a' - 'A';
    int i, c=0, h=0;
    int len = strlen(hwd);

    if (cfg[0] == EMPTY_STRING) return 0;    

    if (hex != 0 && cfg[0] == '0' && cfg[1] == 'x') c = 2;
    if (hex != 0 && hwd[0] == '0' && hwd[1] == 'x') h = 2;
    for (i=0; i<len-h; i++)
	if (cfg[i+c] > hwd[i+h]) {
	    if (cfg[i+c]-case_shift != hwd[i+h])
		break;
	}
	else if (cfg[i+c] < hwd[i+h]) {
	    if (cfg[i+c]+case_shift != hwd[i+h])
		break;
	}

    if (i < strlen(cfg))
	return -1;
    return i;
}

/*
 * List of type aliases
 */
struct interface_alias_match_t {
    hwdesc_interface_class_t class;
    char *name;
    char *alias;
};

struct interface_alias_match_t interface_aliases[] = {
    { iface_class_none, "none", NULL },
    { iface_class_net, "net", 
      "wlan eth lo tr fddi ctc iucv hsi qeth osn escon lcs myri xpnet" },
    { iface_class_eth, "eth", "eth" }, 
    { iface_class_tr, "tr", "tr" }, 
    { iface_class_wlan, "wlan", "wlan" }, 
    { iface_class_block, "block",
      "hd sd sr md dm dasd xpram dcss ida cciss carmel vd vcd fd loop nbd ram umem ad mfm dac960 ed xd" }, 
    { iface_class_partition, "partition", "block" },
    { iface_class_generic, "generic", "sg"},
    { iface_class_sound, "sound", "snd" },
    { iface_class_tty, "tty", "tty" },
    { iface_class_invalid, "invalid", NULL }
};
  
hwdesc_interface_class_t match_interface_type(const char *iface) 
{
    char *s, *e;
    struct interface_alias_match_t *alias_ptr = interface_aliases;

    error_log(LOG_DEBUG,"%s: interface '%s'\n", __FUNCTION__, iface);

    if (!iface || !*iface) return iface_class_none;
    if (*iface == EMPTY_STRING) return iface_class_none;

    while (alias_ptr->class < iface_class_invalid) {
	error_log(LOG_DEBUG,"%s: checking alias '%s'\n",__FUNCTION__, 
		  alias_ptr->name);

	if (!strncmp(alias_ptr->name, iface, strlen(alias_ptr->name))) {
	    error_log(LOG_DEBUG,"%s: found type '%s' for interface '%s'\n",
		      __FUNCTION__, alias_ptr->name, iface);
	    return alias_ptr->class;
	}
	s = alias_ptr->alias;
	while (s && *s) {
	    s++;
	    e = strchr(s,' ');
	    if (!e) e = alias_ptr->alias + strlen(alias_ptr->alias);
	    if (!strncmp(s,iface,(unsigned long)(e - s))) {
		error_log(LOG_DEBUG,"%s: found type '%s' for interface '%s'\n",
			  __FUNCTION__, s, iface);
		return alias_ptr->class;
	    }
	    s = e;
	}
	alias_ptr++;
    }

    error_log(LOG_DEBUG,"%s: invalid type for interface '%s'\n",
	      __FUNCTION__, iface);

    return iface_class_invalid;
}

char *get_interface_name( hwdesc_interface_class_t class )
{
    struct interface_alias_match_t *alias_ptr = interface_aliases;

    while (alias_ptr->name && alias_ptr->class != class)
	alias_ptr++;

    error_log(LOG_DEBUG,"%s: returning type '%s' for class '%x'\n",
	      __FUNCTION__, alias_ptr->name, class);

    return alias_ptr->name;
}

/*
 * List of type aliases
 */
char *type_alias[] = {
    "net wlan eth tr fddi ctc iucv hsi qeth osn escon lcs myri xp",
    "eth qeth hsi lcs",
    "tr qeth lcs",
    "block dasd",
    "scsi_host zfcp",
    "scsi zfcp",
    "ccwgroup lcs ctc ficon qeth osn hsi eth tr",
    "ccw lcs ctc ficon qeth osn hsi dasd zfcp eth tr",
    NULL
};

int match_type(const char *cfg, const char *hwd) {
    char *s;
    char **type_alias_ptr = type_alias;

    error_log(LOG_INFO,"match_type: cfg '%s' hwd '%s'\n", cfg, hwd);

    if ( ! strcmp(cfg, hwd)) return 2;

    if (*hwd == EMPTY_STRING) return 0;
    if (*hwd == '\0') return 0;
    if (*cfg == '\0') return 0;

    while (*type_alias_ptr) {
	if (! strncmp(*type_alias_ptr,hwd,strlen(hwd))) {
	    error_log(LOG_INFO,"match_type: checking aliases '%s'\n", 
		      *type_alias_ptr);
	    s = strchr(*type_alias_ptr,' ');
	    while (s && *s) {
		s++;
		if (!strncmp(s,cfg,strlen(cfg))) {
		    error_log(LOG_INFO,
			      "match_type: hwd '%s' aliased to '%s'\n",
			      hwd,s);
		    return 1;
		}
		s = strchr(s,' ');
	    }
	}
	type_alias_ptr++;
    }

    return -1;
}

void adjust_network_type(hwdesc_t *hwdesc, int ifnum)
{
    /* char subdir[SYSFS_PATH_MAX] = {'\0'}; */
    FILE *procfile;
    char line[256];
    char needle[MAX_HWDESC_PART_LEN] = {'\0'};

    if ( *hwdesc->interface[ifnum] == '\0' )
	return;

    if (!strncmp(hwdesc->interface[ifnum],"eth",3)) {
	hwdesc->interfaceclass[ifnum] = iface_class_eth;
    } else if (!strncmp(hwdesc->interface[ifnum],"tr",2)) {
	hwdesc->interfaceclass[ifnum] = iface_class_tr;
    }
    /* Some drivers do not create /sys/class/net/iface/wireless, therefore we
     * look now in /proc/net/wireless for the interface. I leave the old code
     * here; maybe we have to reactivate it for some strange reason. - zoz */
    /*    snprintf(subdir, SYSFS_PATH_MAX, "%s/wireless",
	     hwdesc->interfacepath[ifnum]);
    if ( ! sysfs_path_is_dir(subdir)) {
	hwdesc->interfaceclass[ifnum] = iface_class_wlan;
	return;
	} */
    procfile = fopen("/proc/net/wireless", "r");
    if (!procfile)
	return;
    snprintf(needle, MAX_HWDESC_PART_LEN -1 , "%s:",
	     hwdesc->interface[ifnum]);
    while (fgets(line, 256, procfile) != NULL) {
	if (strstr(line, needle) != NULL) {
	    hwdesc->interfaceclass[ifnum] = iface_class_wlan;
	    break;
	}
    }
    fclose(procfile);
}

void check_or_add_interface(hwdesc_t *hwdesc, char *interfacepath, 
			    hwdesc_interface_class_t this_iface_class)
{
    char interfacename[SYSFS_PATH_MAX] = {'\0'};
    int ifnum = 0;
    int fixup_iface = -1;

    sysfs_get_name_from_path(interfacepath,
			     interfacename, SYSFS_PATH_MAX);

    error_log(LOG_DEBUG,"Test iface '%s', path '%s' class %d\n", 
	      interfacename, interfacepath, this_iface_class);

    /* Check existing interfaces */
    while (ifnum < hwdesc->interface_num) {
	error_log(LOG_DEBUG,"Check iface %d (%s), path '%s' type '%d'\n",
		  ifnum, hwdesc->interface[ifnum],
		  hwdesc->interfacepath[ifnum],
		  hwdesc->interfaceclass[ifnum] );

	if (match_interface_class(hwdesc->interfaceclass[ifnum],
				  this_iface_class) ||
	    hwdesc->interfaceclass[ifnum] == iface_class_invalid ) {
	    /* Fixup interfacepath */
	    if ( *hwdesc->interfacepath[ifnum] != '\0' &&
		 !strcmp(hwdesc->interfacepath[ifnum], interfacepath) ) {
		fixup_iface = ifnum;
	    }
	    /* Fixup interface name */
	    if ( *hwdesc->interface[ifnum] != '\0' &&
		 !strcmp(hwdesc->interface[ifnum], interfacename) ) {
		fixup_iface = ifnum;
	    }
	    if (fixup_iface == ifnum) {
		/* Fixup interface */
		if (*hwdesc->interface[ifnum] == '\0') {
		    strcpy(hwdesc->interface[ifnum], interfacename);
		}
		if (*hwdesc->interfacepath[ifnum] == '\0') {
		    strcpy(hwdesc->interfacepath[ifnum], interfacepath);
		}
		hwdesc->interfaceclass[ifnum] = this_iface_class;
		if (this_iface_class == iface_class_net)
		    adjust_network_type(hwdesc, ifnum);

		strcpy(hwdesc->interfacetype[ifnum], 
		       get_interface_name(this_iface_class));

		error_log(LOG_DEBUG,"Fixup iface '%s', path '%s' type '%s'\n",
			  hwdesc->interface[ifnum],
			  hwdesc->interfacepath[ifnum],
			  hwdesc->interfacetype[ifnum] );
	    }
	}
	ifnum++;
    }
    /* Add additional interface if required */
    if (fixup_iface == -1) {
	hwdesc->interfaceclass[ifnum] = this_iface_class;
	strcpy(hwdesc->interface[ifnum], interfacename);
	strcpy(hwdesc->interfacepath[ifnum], interfacepath);

	if (this_iface_class == iface_class_net)
	    adjust_network_type(hwdesc, ifnum);

	strcpy(hwdesc->interfacetype[ifnum], 
	       get_interface_name(this_iface_class));

	error_log(LOG_DEBUG,"Add iface '%s', path '%s' type '%s'\n",
		  hwdesc->interface[ifnum],
		  hwdesc->interfacepath[ifnum],
		  hwdesc->interfacetype[ifnum] );

	hwdesc->interface_num++;
    }
}

int get_configuration(hwdesc_t *hwdesc,
		      const char *arg_dirname, const char* arg_prefix) {
    DIR *directory;
    struct dirent *file;
    char *s;
    int i, k, n;
    int len, bonus;
    hwdesc_type_t best_type = invalid;
    int level = 0, best_level = 0;
    char best_name[MAX_HWDESC_LEN] = {'\0'};
    char dirname[MAX_HWDESC_LEN] = {'\0'};
    char prefix[MAX_PREFIX_LEN] = {'\0'};
    hwdesc_t *config;
    int hex = 0;

    error_log(LOG_DEBUG,"%s\n",__FUNCTION__);

    config = malloc(sizeof(hwdesc_t));
    if (config == NULL)
	return -1;

    /* Why compare with class/net ? -hare */
    /* Because sysconfig/network is for network interfaces only. -zoz */
    if (arg_dirname != NULL && *arg_dirname != '\0')
	strncpy(dirname, arg_dirname, MAX_HWDESC_LEN);
    else if (hwdesc->hwdesc_type == interface &&
	     strstr(hwdesc->interfacepath[0], "class/net"))
	strcpy(dirname, "/etc/sysconfig/network");
    else
	strcpy(dirname, "/etc/sysconfig/hardware");

    if (arg_prefix !=NULL && *arg_prefix != '\0')
	strncpy(prefix, arg_prefix, MAX_PREFIX_LEN);
    else if (hwdesc->hwdesc_type == interface &&
	     strstr(hwdesc->interfacepath[0], "class/net"))
	strcpy(prefix, "ifcfg-");
    else
	strcpy(prefix, "hwcfg-");

    /* If type or last busname is pci or pcmcia then we skip '0x' at the
     * beginning of any part of vp id before comparing. This will be done
     * by setting 'hex' to 1, which is used in calls to 'match_id()'. */
    if ( ! strcmp(hwdesc->devtype, "pci") ) hex = 1;
    if ( ! strcmp(hwdesc->devtype, "pcmcia") ) hex = 1;
    if (hwdesc->bus_num > 0) {
	if ( ! strcmp(hwdesc->busname[hwdesc->bus_num-1], "pci")) hex = 1;
	if ( ! strcmp(hwdesc->busname[hwdesc->bus_num-1], "pcmcia")) hex = 1;
    }

    /* Every filename that starts with ifcfg- or hwcfg- is without that prefix
     * a configuration name. Now we put it in a struct hwdesc_t named 'config'
     * and split it into its components. Then compare it with the hardware
     * description in 'hwdesc'. If there was a match, write the type of the
     * configuration name, the matching level and the name itself into a struct
     * match_t in the list 'match_head'. Later we can easily get the
     * configuration name that matched best. */
    directory = opendir(dirname);
    if (directory == NULL) {
	error_log(LOG_ERR, "Cannot open directory %s\n", dirname);
	return -1;
    }
    while ((file = readdir(directory)) != NULL) {
	/* Drop all files that do not start with 'prefix' */
	s = strstr(file->d_name, prefix);
	if (s != file->d_name)
	    continue;
	strncpy(config->hwdesc, file->d_name + strlen(prefix), MAX_HWDESC_LEN);
	split_hwdesc(config);
//	print_hwdesc(config, file->d_name, 0);

	/* Comparison depends on the type of 'config'. If we already have a
	 * match with a better type we don't need to compare at all. */
	if (config->hwdesc_type > best_type)
	    continue;
	level = 0;
	switch (config->hwdesc_type) {
	    case id:
		/* id must only match partially, i.e. config->id might be
		 * shorter as hwdesc->id, but whole config->id must match. */
		level += match_id(config->id, hwdesc->id, 1, 0);
		break;
	    case vpid:
		/* like id, but in five parts: a-aa does not match aa-a */

		len = match_id(config->vendorid, hwdesc->vendorid, 1, hex);
		if (len < 0) { level = 0; break; }
		level += len;
//		printf("%-30s\t %s\t %s\t %d\t %d\n", file->d_name,
//		       config->vendorid, hwdesc->vendorid, level, len);

		len = match_id(config->productid, hwdesc->productid, 1, hex);
		if (len < 0) { level = 0; break; }
		level += len;
//		printf("%-30s\t %s\t %s\t %d\t %d\n", file->d_name,
//		       config->productid, hwdesc->productid, level, len);

		len = match_id(config->vendorsubid, hwdesc->vendorsubid,
			       1, hex);
		if (len < 0) { level = 0; break; }
		level += len;
//		printf("%-30s\t %s\t %s\t %d\t %d\n", file->d_name,
//		       config->vendorsubid, hwdesc->vendorsubid, level, len);

		len = match_id(config->productsubid, hwdesc->productsubid,
			       1, hex);
		if (len < 0) { level = 0; break; }
		level += len;
//		printf("%-30s\t %s\t %s\t %d\t %d\n", file->d_name,
//		       config->productsubid, hwdesc->productsubid, level, len);

		len = match_id(config->vpidrev, hwdesc->vpidrev, 1, hex);
		if (len < 0) { level = 0; break; }
		level += len;
//		printf("%-30s\t %s\t %s\t %d\t %d\n", file->d_name,
//		       config->vpidrev, hwdesc->vpidrev, level, len);

		break;
	    case bus:
		/* busid must only match partially, i.e. config->busid might be
		 * shorter as hwdesc->busid, but whole config->busid must
		 * match. busname must always match completely. */
		bonus = 0;
		for (n=0; n<hwdesc->bus_num; n++) {

		    for (i=0; i<config->bus_num && i<hwdesc->bus_num-n; i++) {
			len = strlen(hwdesc->busid[i+n]);
			if (*config->busname[i] == '\0')
			    break;

			error_log(LOG_DEBUG,"bus %d: %s len %d hwdesc %s\n", i,
				  config->busname[i], len, hwdesc->busname[i+n]);

			/* Bus names must match */
			if (strcmp(config->busname[i], hwdesc->busname[i+n])) {
			    level = 0;
			    break;
			} else
			    level += 1 + len;
			/* Bus ids may be empty */
			if (*config->busid[i] == '\0' || 
			    *hwdesc->busid[i+n] == '\0')
			    break;
			for (k=0; k<len; k++)
			    if (config->busid[i][k] != hwdesc->busid[i+n][k])
				break;
			if (k < strlen(config->busid[i])) {
			    level = 0;
			    break;
			} else
			    level += k;
		    }
		    if (level != 0) {
			level += bonus;
			break;
		    }
		    bonus += len * 2;
		}
		break;
	    case devpath:
		if (! strcmp(config->devpath, hwdesc->devpath))
		    level++;
		break;
	    case interface:
		// only test against first hwdesc for ib
		if(!strncmp("ib",hwdesc->interface[0],2)) {
		    if (! strcmp(config->interface[0], hwdesc->interface[0]))
			level++;
		}
		else {
		    for (i=0; i<hwdesc->interface_num; i++) {
			if (! strcmp(config->interface[0],
			             hwdesc->interface[i]))
			    level++;
		    }
		}
		break;
	    case type:
		if (0 <= match_type(config->devtype, hwdesc->devtype))
		    level++;
		for (i=0; i<hwdesc->interface_num; i++)
		  {
		    // if (! strcmp(config->devtype,
		    // hwdesc->interfacetype[i]))
		      if (0 <= match_type(config->devtype, 
					  hwdesc->interfacetype[i]))
		      {
			level++;
		      }
		  }
		break;
	    case invalid:
	    case none:
		break;
	}

	error_log(LOG_DEBUG,"name: %-35s level: %2d type: %d  best_type: %d\n",
		  file->d_name, level, config->hwdesc_type, best_type);

	if (level <= 0)
	    continue;
	level *= 2;
	len = match_type(config->devtype, hwdesc->devtype);
	if (len < 0)
	    continue;
	if (len > 0)
	    level++;
//	printf("name: %-35s level: %2d   type: %d   best_type: %d\n",
//	       file->d_name, level, config->hwdesc_type, best_type);
////	       "", level, config->hwdesc_type, best_type);
	if (config->hwdesc_type == best_type && level < best_level)
	    continue;

	best_type  = config->hwdesc_type;
	best_level = level;
	strcpy(best_name, config->hwdesc);

    }
    closedir(directory);

    if (*best_name != '\0')
	/* Don't know what is best to do if there is already a maximum number
	 * of configs and if it ever happens in real life. So let's return an
	 * error for now. */
	if (hwdesc->config_num < MAX_INTERFACE_NUM) {
	    strcpy(hwdesc->config[hwdesc->config_num++], best_name);
	    return 0;
	}

    return 1;
}

int get_all_interfaces(const char *classname, const char *configname,
		       char *interfaces) {
    int r = 0;
    char mnt_path[SYSFS_PATH_MAX] = {'\0'};
    struct dlist *ifacelist = NULL;
    struct sysfs_class *sfc_class;
    struct sysfs_class_device *sfcd_iface;
    hwdesc_t *hwdesc;

    error_log(LOG_DEBUG,"%s\n",__FUNCTION__);

    hwdesc = malloc(sizeof(hwdesc_t));
    if (hwdesc == NULL)
	return -1;

    memset(interfaces, 0, sizeof(interfaces));


    if (sysfs_get_mnt_path(mnt_path, SYSFS_PATH_MAX)) {
	error_log(LOG_ERR, "Cannot get sysfs mnt path\n");
	return 401;
    }       

    sfc_class = sysfs_open_class(classname);
    if (sfc_class == NULL) {
	error_log(LOG_WARNING, "Could not open class %s\n", classname);
	return -1;
    }

    ifacelist = sysfs_get_class_devices(sfc_class);
    if (ifacelist == NULL) {
	error_log(LOG_INFO, "No interfaces in class %s\n",
		  classname);
	return 0;
    }

    dlist_for_each_data(ifacelist, sfcd_iface, struct sysfs_class_device) {
	memset(hwdesc, 0, sizeof(hwdesc_t));
	hwdesc->interface_num=1;
	strcpy(hwdesc->interface[0], sfcd_iface->name);
	hwdesc->hwdesc_type=interface;
	complete_hwdesc_sysfs(hwdesc);
	get_configuration(hwdesc, "/etc/sysconfig/network", "ifcfg-");
//	print_hwdesc(hwdesc, sfcd_iface->name, 0);
//	printf("interface: '%s'  config: '%s' matchconfig: '%s'\n",
//	       sfcd_iface->name, hwdesc->config[0], configname);
	if (*configname != '\0' && strcmp(hwdesc->config[0], configname))
		continue;
	if (*interfaces != '\0')
	    strcat(interfaces, " ");
	strcat(interfaces, sfcd_iface->name);
	r++;
    }
    sysfs_close_class(sfc_class);


    return r;
}

int look_pcmcia_stab(const char *look, char *found, const char *compare,
		     int n_look, int n_found, int n_compare){
    FILE *stab;
    char line[256];
    char field[8][32];

    error_log(LOG_DEBUG,"%s\n",__FUNCTION__);

    stab = fopen("/var/lib/pcmcia/stab", "r");
    if (stab == NULL) {
	error_log(LOG_WARNING, "No /var/lib/pcmcia/stab\n");
	return -1;
    }
    while (1) {
	if (fgets(line, 256, stab) == NULL)
	    break;
//	printf("%s", line);
	if ( ! sscanf(line, "%32[0-9] %32s %32s %32s %32s", field[0],
		      field[1], field[2], field[3], field[4]))
	    continue;
	error_log(LOG_DEBUG,"PCMCIA stab: slot %s type %s interface %s\n", 
		  field[0], field[1], field[4]);
	if (strcmp(field[n_look], look))
	    continue;
	if (*compare != '\0' && strcmp(field[n_compare], compare))
	    continue;
//	printf("daisses\n");
	strcpy(found, field[n_found]);
	error_log(LOG_DEBUG,"-> found %s\n", found);
	return 0;
    }

    return 1;
}

int look_smpppd_info(char *name){

    char filename[MAX_DEVPATH_LEN] = {'\0'};
    char line[256];
    char field[8][32];
    FILE *info;

    error_log(LOG_DEBUG,"%s: name %s\n",__FUNCTION__,name);

//    printf("look_smpppd_info: %s\n", name);
    sprintf(filename, "/var/lib/smpppd/ifcfg-%s.info", name);
    info = fopen(filename, "r");
    if (info == NULL) {
	error_log(LOG_WARNING, "No %s\n", filename);
//	printf("No %s\n", filename);
	return -1;
    }
    while (1) {
	if (fgets(line, 256, info) == NULL)
	    break;
//	printf("%s", line);
	if ( ! sscanf(line, "%32s %32s %32s", field[0],
		      field[1], field[2]))
	    continue;
//	printf("Name: %s  Value: %s   Rest: %s\n", field[0],
//	       field[1], field[2]);
	if (strcmp(field[0], "interface:"))
	    continue;
	strcpy(name, field[1]);
	return 0;
    }

    return 1;
}

int complete_hwdesc_sysfs(hwdesc_t *hwdesc) {
    /* Here we try to get some more information from sysfs. We try to get
     * hwdesc->devicepath and hwdesc->interfacepath.
     * Then we can get much more information out of sysfs.
     * That's what we have to do:                                   implemented
     *  1) get interfacepath and/or devicepath from devpath          yes
     *  2) get interfacepath from interface                          yes
     *  3) get interfacepath and interface from id                   partial
     *     only for class net
     *  4) get devicepath from interfacepath                         yes
     *  5) get devicepath from bus info                              yes
     *  6) get devicepath from id                                    no
     *  7) get devicepath from vpid                                  no
     *  8) get devicepath from type                                  no
     * 10) get interfacepath from devicepath                         yes
     * 11) get buspaths from devicepath                              partial
     *     currently we just set bus info but don't store buspaths
     * 21) get businfo from any path                                 partial
     *     see 11)
     * 22) get id from any path                                      partial
     *     from interfacepath, attribute 'address'
     *     from devicepath, attribute 'serial'
     * 23) get vpid from devicepath                                  partial
     *     only for pci
     * 24) get interface from interfacepath                          yes
     * 25) get type from any path                                    no
     * 30) get driver name from any path                             yes
     */
    int i;
    int is_id = 0;
    int best_id;
    int ifnum;
    char *interface_paths[] = {
	"class/net",
	SYSFS_BLOCK_NAME,
	NULL
    };
    char *device_paths[] = {
	SYSFS_DEVICES_NAME,
	NULL
    };
    char mnt_path[SYSFS_PATH_MAX] = {'\0'};

    char subdir[SYSFS_PATH_MAX] = {'\0'};
    char devicename[SYSFS_PATH_MAX] = {'\0'};
    char devicepath[SYSFS_PATH_MAX] = {'\0'};
    char devicelink[SYSFS_PATH_MAX] = {'\0'};
    char interfacepath[SYSFS_PATH_MAX] = {'\0'};
    char pcmcia_info[MAX_HWDESC_PART_LEN] = {'\0'};
    char *s, *t;
    char *lp = NULL, *rp = NULL;
    char *busname = NULL;
    char *devname = NULL;
    struct dlist *devlist = NULL;
    struct dlist *ifacelist = NULL;
    struct sysfs_directory *sfd_dir = NULL;
    struct sysfs_attribute *sfa_id = NULL;
    struct sysfs_class *sfc_class;
    struct sysfs_class_device *sfcd_iface;
    struct sysfs_class_device *sfcd_best_iface;
    struct sysfs_device *sfdv_device;


    error_log(LOG_DEBUG,"%s\n",__FUNCTION__);

    if (sysfs_get_mnt_path(mnt_path, SYSFS_PATH_MAX)) {
	error_log(LOG_ERR, "Cannot get sysfs mnt path\n");
	return 401;
    }       
//    strcpy(mnt_path, "/home/zoz/sysconfig/sysfs.examples/usb-nic/sys");
//    strcpy(mnt_path, "/sys");
//    printf("NOTE:\nNOTE: hardcoded sysfs mount point\nNOTE: %s\nNOTE:\n",
//	   mnt_path);



    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
     * 1) get devicepath from devpath
     * ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
     * If we already got a devpath, look if it's a path to a device
     * (/sys/device) or to an interface (/sys/class or /sys/block) and set
     * devicepath resp. interfacepath if not already set.
     */
    error_log(LOG_DEBUG,"Step 1: devicepath from devpath\n");

    if (*hwdesc->devpath != '\0') {
	/* devpath is usually meant relative to 'mnt_path' even if it starts
	 * with a '/'. Therefore create a valid absolute path first. */
	s = strstr(hwdesc->devpath, mnt_path);
	if (s != hwdesc->devpath) {
	    s = index(hwdesc->devpath, '/');
	    if (s == hwdesc->devpath)
		snprintf(devicepath, SYSFS_PATH_MAX, "%s%s", mnt_path,
			 hwdesc->devpath);
	    else
		snprintf(devicepath, SYSFS_PATH_MAX, "%s/%s", mnt_path,
			 hwdesc->devpath);
	}
	else
	    strcpy(devicepath, hwdesc->devpath);

	error_log(LOG_DEBUG,"devpath = %s\n", devicepath);

	/* Check for links first */
	strcpy(subdir, devicepath);
	if (!sysfs_get_link(subdir, hwdesc->devicepath,
			    SYSFS_PATH_MAX) != 0) {
	    strcpy(devicepath, hwdesc->devicepath);
	    hwdesc->devicepath[0] = '\0';
	}

	if (*hwdesc->devicepath == '\0')
	    /* Check whether it's a real device */
	    for (i=0; device_paths[i] != NULL; i++) {
		snprintf(subdir, SYSFS_PATH_MAX, "%s/%s",
			 mnt_path, device_paths[i]);
		s = strstr(devicepath, subdir);
		/* It must match at the start of the devpath */
		if (s != devicepath)
		    continue;
		strcpy(hwdesc->devicepath, devicepath);
	    }
	if (*hwdesc->devicepath == '\0') {
	    /* Or maybe it's a device link */
	    sfcd_iface = sysfs_open_class_device_path(devicepath);
	    *devicepath = '\0';
	    if (sfcd_iface != NULL) {
		sfdv_device = sysfs_get_classdev_device(sfcd_iface);
		if (sfdv_device)
		    strcpy(devicepath, sfdv_device->path);
		sfdv_device = NULL;
	    }
	    sysfs_close_class_device(sfcd_iface);
	    sfcd_iface = NULL;

	    if (!*devicepath) {
		error_log(LOG_DEBUG,"invalid device path: %s\n", 
			  subdir);
	    } else {
		strcpy(hwdesc->devicepath, devicepath);
	    }
	}
	/*
	 * ccwgroup hack:
	 * Even though the device appears as a grouped device
	 * we would rather access it via the first real device
	 */
	if (*hwdesc->devicepath != '\0') {
	    snprintf(subdir, SYSFS_PATH_MAX,"%s/cdev0",
		    hwdesc->devicepath);
	    if (!sysfs_get_link(subdir, devicepath,
				SYSFS_PATH_MAX) != 0) {
		strcpy(hwdesc->devicepath, devicepath);
	    }
	}
    }
    error_log(LOG_DEBUG,"-> devicepath %s\n", hwdesc->devicepath);

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
     * 2) get interfacepath from interface
     * ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ 
     * If we still have no interfacepath, but have interfaces, then lets look
     * for them in sysfs.
     * We have to search at class/<any>/, block/ and block/<any>/
     */
    error_log(LOG_DEBUG,"Step 2: interfacepath from interface (%d interfaces)\n",
	   hwdesc->interface_num);

    ifnum = 0;
    while (ifnum < hwdesc->interface_num) {
	if (*hwdesc->interfacepath[ifnum] == '\0') {
	    /* 
	     * Two passes:
	     * First try to access the interface directly
	     * failing that scan all subdirs of al interfaces
	     * (which might be quite slow, but we'll have to do
	     * it if we don't get anything in the first pass.
	     */

	    /* Sometimes interface names are 'device files' (/dev/hda1),
	     * but in sysfs they are just called by their filenames.
	     * Therefore we strip a possible path. */
	    if (sysfs_get_name_from_path(hwdesc->interface[ifnum], subdir,
					 SYSFS_PATH_MAX)) {
		strcpy(subdir, hwdesc->interface[ifnum]);
	    }

	    /* First pass: Access interface directly */
	    for (i=0; interface_paths[i] != NULL && 
		     *hwdesc->interfacepath[ifnum] == '\0'; i++) {
		snprintf(interfacepath, SYSFS_PATH_MAX, "%s/%s/%s",
			 mnt_path, interface_paths[i],subdir);
		error_log(LOG_DEBUG,"Checking interface path %s\n",
			  interfacepath);

		sfcd_iface = sysfs_open_class_device_path(interfacepath);
		if (sfcd_iface == NULL) {
		    error_log(LOG_DEBUG,"invalid interface path: %s\n", 
			      interfacepath);
		    continue;
		}
		strcpy(hwdesc->interfacepath[ifnum], sfcd_iface->path);
		hwdesc->interfaceclass[ifnum] = match_interface_type(hwdesc->interface[ifnum]);
		sysfs_close_class_device(sfcd_iface);
		sfcd_iface = NULL;
	    }
	    /* Second pass: scan block interfaces for partitions */
	    if (*hwdesc->interfacepath[ifnum] == '\0') {
		/* Don't try to be overly clever here.
		 * Better to scan all interfaces than to not find the
		 * correct one ...
		 */
		error_log(LOG_DEBUG,"Scanning for interfaces:\n"); 
		sfc_class = sysfs_open_class("block");
		if (sfc_class) {
		    ifacelist = sysfs_get_class_devices(sfc_class);

		    dlist_for_each_data(ifacelist, sfcd_iface, 
					struct sysfs_class_device) {
			sprintf(subdir,"%s/%s",sfcd_iface->path,
				hwdesc->interface[ifnum]);
			error_log(LOG_DEBUG,"Checking interface path %s\n", 
				  subdir);
			sfd_dir = sysfs_open_directory(subdir);
			if (sfd_dir) {
			    error_log(LOG_DEBUG,"Found interface path '%s'\n", 
				      sfd_dir->path);
			    strcpy(hwdesc->interfacepath[ifnum], 
				   sfd_dir->path);
			    hwdesc->interfaceclass[ifnum] = iface_class_partition;
			    sysfs_close_directory(sfd_dir);
			    break;
			}
		    }
		    sysfs_close_class(sfc_class);
		}
	    }
	}
	error_log(LOG_DEBUG,"-> interfacepath %s\n", 
		  hwdesc->interfacepath[ifnum]);
	ifnum++;
    }

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
     * 3) get interfacepath and interface from id
     * ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
     * We currently look only for addresses of network interfaces
     */
    error_log(LOG_DEBUG,"Step 3: interfacepath and interface from id\n");

    if (*hwdesc->id != '\0') {
	sfc_class = sysfs_open_class("net");
	if (sfc_class == NULL) 
	    error_log(LOG_ERR, "Cannot open class net\n");
	else {
	    ifacelist = sysfs_get_class_devices(sfc_class);
	    if (ifacelist == NULL)
		error_log(LOG_INFO, "No interfaces in class net");
	    else {
		best_id = 0;
		sfcd_best_iface = NULL;
		dlist_for_each_data(ifacelist, sfcd_iface,
				    struct sysfs_class_device) {
		    error_log(LOG_DEBUG,"interface: %s --- ", sfcd_iface->name);
		    sfa_id = sysfs_get_classdev_attr(sfcd_iface, "address");
		    if (sfa_id == NULL) {
			error_log(LOG_INFO, "Interface %s has no attribute"
				  "'address'\n", sfcd_iface->name);
			error_log(LOG_DEBUG,"no address\n");
		    } else {
			is_id = match_id(hwdesc->id, sfa_id->value, 1, 0);
			error_log(LOG_DEBUG,"address: %s is_id=%d\n", 
				  sfa_id->value, is_id);
		    }
		    if (is_id > best_id) {
			best_id = is_id;
			sfcd_best_iface = sfcd_iface;
		    }
		}
		if (sfcd_best_iface != NULL) {
		    error_log(LOG_DEBUG,"best interface: %s\n", 
			      sfcd_best_iface->name);
		    check_or_add_interface(hwdesc, sfcd_best_iface->path,
					   iface_class_net);
		}
	    }
	    ifacelist = NULL;
	    sysfs_close_class(sfc_class);
	}
    }

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
     * 4) get devicepath from interfacepath
     * ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ 
     * If we still have no devicepath, but an interfacepath, then lets look
     * for the device link in this directory or, if interface is a disk
     * partition, in its parent directory.
     *
     * Interfaces on PC-cards have no device link. In this case we look in
     * a table from cardmgr for the pcmcia socket. These have a device link.
     * But later in 'get businfo ...' we have to add pcmcia-<socketnr>.
     */
    error_log(LOG_DEBUG,"Step 4: devicepath from interfacepath\n");

    if (*hwdesc->devicepath == '\0') {
	ifnum = 0;
	while (ifnum < hwdesc->interface_num) {
	    error_log(LOG_DEBUG,"check device %s class %d\n",
		      hwdesc->interface[ifnum], hwdesc->interfaceclass[ifnum]);
	    if (*hwdesc->interfacepath[ifnum] != '\0') {
		snprintf(subdir, SYSFS_PATH_MAX, "%s/%s",
			 hwdesc->interfacepath[ifnum], "device");
		sysfs_get_link(subdir, hwdesc->devicepath, SYSFS_PATH_MAX);

		if (*hwdesc->devicepath == '\0' &&
		    hwdesc->interfaceclass[ifnum] == iface_class_block ) {
		    strcpy(subdir,hwdesc->interfacepath[ifnum]);
		    lp = strrchr(subdir, '/');
		    if (lp) {
			*lp = '\0';
			strcat(subdir,"/device");
			error_log(LOG_DEBUG,"Check for partition on '%s'\n",
				  subdir);
			sysfs_get_link(subdir, hwdesc->devicepath, 
				       SYSFS_PATH_MAX);
			hwdesc->interfaceclass[ifnum] = iface_class_partition;
		    }
		}
	    }
	    if (*hwdesc->devicepath == '\0') {
		if ( ! look_pcmcia_stab(hwdesc->interface[ifnum], 
					pcmcia_info, "", 4, 0, 0)) {
		    error_log(LOG_DEBUG,"Interface %s has slot %s\n", 
			      hwdesc->interface[ifnum], pcmcia_info);
		    snprintf(subdir, SYSFS_PATH_MAX,
			     "%s/%s/pcmcia_socket/pcmcia_socket%s/device",
			     mnt_path, SYSFS_CLASS_NAME, pcmcia_info);
		    sysfs_get_link(subdir, hwdesc->devicepath, SYSFS_PATH_MAX);
		}
	    }
	    ifnum++;
	}
	error_log(LOG_DEBUG,"-> devicepath %s\n", hwdesc->devicepath);
    } else {
	error_log(LOG_DEBUG,"-> devicepath (already existing) %s\n", 
		  hwdesc->devicepath);
    }

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
     * 5) get devicepath from bus info
     * ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
     */
    error_log(LOG_DEBUG,"Step 5: devicepath from bus info\n");

    if (*hwdesc->devicepath == '\0' && hwdesc->bus_num > 0) {
	snprintf(subdir, SYSFS_PATH_MAX, "%s/%s/%s/devices/%s",
		 mnt_path, SYSFS_BUS_NAME,
		 hwdesc->busname[hwdesc->bus_num - 1],
		 hwdesc->busid[hwdesc->bus_num - 1]);
	sysfs_get_link(subdir, hwdesc->devicepath, SYSFS_PATH_MAX);
	error_log(LOG_DEBUG,"-> devicepath %s\n", hwdesc->devicepath);
    }
    if (*hwdesc->devicepath == '\0' && hwdesc->bus_num > 0) {
	if ( ! strcmp(hwdesc->busname[hwdesc->bus_num - 1], "pcmcia")) {
	    strcpy(pcmcia_info, hwdesc->busid[hwdesc->bus_num -1]);
//	    printf("Look for pcmcia slot %s\n", pcmcia_info);
	    snprintf(subdir, SYSFS_PATH_MAX,
		     "%s/%s/pcmcia_socket/pcmcia_socket%s/device",
		     mnt_path, SYSFS_CLASS_NAME, pcmcia_info);
	    sysfs_get_link(subdir, hwdesc->devicepath, SYSFS_PATH_MAX);
	}
	/* If given bus is pcmcia, look if there is another device plugged into
	 * the pcmcia slot and take this, if not explicitly asked for pcmcia
	 * bridges. */
	if ( ! strcmp(hwdesc->busname[hwdesc->bus_num - 1], "pcmcia") &&
	     strcmp(hwdesc->devtype, "pcmcia")) {
	    sfd_dir = sysfs_open_directory(hwdesc->devicepath);
	    if (sfd_dir != NULL) {
		devlist = sysfs_get_dir_subdirs(sfd_dir);
		if (devlist != NULL) {
		    dlist_for_each_data(devlist, devname, char) {
//			printf("devname: %s\n", devname);
			snprintf(subdir, SYSFS_PATH_MAX, "%s/%s",
				 sfd_dir->path, devname);
//			printf("subdir: %s\n", subdir);
			sfdv_device = sysfs_open_device_path(subdir);
			if (sfdv_device != NULL) {
			    strncpy(hwdesc->devicepath, sfdv_device->path,
				    MAX_DEVPATH_LEN);
			    sysfs_close_device(sfdv_device);
				break;
			}
		    }
		}
		sysfs_close_directory(sfd_dir);
	    }
	}
	error_log(LOG_DEBUG,"-> devicepath (pcmcia) %s\n", hwdesc->devicepath);
    }
 
    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
     * 6) get devicepath from id
     * ^^^^^^^^^^^^^^^^^^^^^^^^^
     */
    error_log(LOG_DEBUG,"Step 6: devicepath from id (N/A)\n");
 
    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
     * 7) get devicepath from vpid
     * ^^^^^^^^^^^^^^^^^^^^^^^^^^^
     */
    error_log(LOG_DEBUG,"Step 7: devicepath from vpid (N/A)\n");
 
    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
     * 8) get devicepath from type
     * ^^^^^^^^^^^^^^^^^^^^^^^^^^^
     */
    error_log(LOG_DEBUG,"Step 8: devicepath from type (N/A)\n");

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
     *
     * Start of reverse mapping.
     * From this point we should have a devicepath.
     * Start filling out missing information.
     */
    if (*hwdesc->devicepath == '\0') {
	error_log(LOG_INFO, "Cannot identify device (no devicepath)\n");
    } else {

	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
	 * 10) get interfacepath from devicepath
	 * ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
	 * Currently, there are only 2 interface types: block and net.
	 * block interfaces have a proper back-link in symlink 'block'.
	 * So the only class worth scanning is 'net'.
	 */
	error_log(LOG_DEBUG,"Step 10: interfacepath from devicepath (path %s)\n",
		  hwdesc->devicepath);
	
	/* First check whether existing interface paths match */
	ifnum = 0;
	while (ifnum  < hwdesc->interface_num) {
	    error_log(LOG_DEBUG,"Interface '%s': ", hwdesc->interface[ifnum]);
	    
	    if (*hwdesc->interfacepath[ifnum] != '\0') {
		sprintf(subdir,"%s/device",hwdesc->interfacepath[ifnum]);
		if (!sysfs_get_link(subdir, devicelink,
				    SYSFS_PATH_MAX) != 0) {
		    error_log(LOG_DEBUG,"device link %s ", devicelink);
		    if (strcmp(hwdesc->devicepath, devicelink)) {
			/* Interface path does not match, resetting */
			error_log(LOG_DEBUG,"-> No match, resetting !\n");
			*hwdesc->interfacepath[ifnum] = '\0';
			hwdesc->interfaceclass[ifnum] = iface_class_invalid;
		    } else {
			error_log(LOG_DEBUG,"-> Ok.\n");
		    }
		}
	    }
	    ifnum++;
	}
	
	/* Now find any remaining or invalid interfaces */
	
	/* Check for block devices */
	sprintf(devicelink,"%s/block",hwdesc->devicepath);
	error_log(LOG_DEBUG,"Trying device link %s\n", devicelink);
	if (sysfs_get_link(devicelink, interfacepath,
			   SYSFS_PATH_MAX) == 0) {
	    error_log(LOG_DEBUG,"Found block device %s\n",
		      interfacepath);
	    check_or_add_interface(hwdesc, interfacepath, iface_class_block);
	}
	
	/* Check for generic devices */
	sprintf(devicelink,"%s/generic",hwdesc->devicepath);
	error_log(LOG_DEBUG,"Trying device link %s\n", devicelink);
	if (sysfs_get_link(devicelink, interfacepath,
			   SYSFS_PATH_MAX) == 0) {
	    error_log(LOG_DEBUG,"Found generic device %s\n",
		      interfacepath);
	    check_or_add_interface(hwdesc, interfacepath, iface_class_generic);
	}
	
	/* Check for S/390 grouped devices */
	sprintf(devicelink,"%s/group_device", hwdesc->devicepath);
	if (sysfs_get_link(devicelink, interfacepath,
			   SYSFS_PATH_MAX) != 0) {
	    strcpy(devicelink, hwdesc->devicepath);
	} else {
	    strcpy(devicelink, interfacepath);
	}

	/* Check for network devices */
	error_log(LOG_DEBUG,"checking network device:  path %s\n", 
		  devicelink);
	sfd_dir = sysfs_open_directory(devicelink);
	if (sfd_dir != NULL) {
	    devlist = sysfs_get_dir_links(sfd_dir);
	    if (devlist != NULL) {
		dlist_for_each_data(devlist, devname, char) {
		    error_log(LOG_DEBUG, "devname: %s\n", devname);
		    if (!strncmp(devname,"net:",4)) {
			/* network interface found */
			snprintf(subdir, SYSFS_PATH_MAX, "%s/%s",
				 sfd_dir->path, devname);
			error_log(LOG_DEBUG, "devlink: %s\n", devicelink);
			if (sysfs_get_link(subdir, devicelink,SYSFS_PATH_MAX) == 0) {
			    check_or_add_interface(hwdesc, devicelink, 
						   iface_class_net);
			}
		    }
		}
	    }
	    sysfs_close_directory(sfd_dir);
	}
	
	/* Check for pcmcia devices */
	error_log(LOG_DEBUG,"checking pcmcia device:  path %s\n",
		  hwdesc->devicepath);
	/* If last busname is "pcmcia" and devtype is not "pcmcia" then
	 * look if this pcmcia_socket hosts another interface and take this. */
	if ( ! strcmp(hwdesc->busname[hwdesc->bus_num - 1], "pcmcia") &&
	     strcmp(hwdesc->devtype, "pcmcia")) {
	    error_log(LOG_DEBUG,"checking pcmcia device:  path %s\n", 
		      hwdesc->devicepath);
	    
	    /* FIXME: look also for other devtypes, not only network */
	    if ( ! look_pcmcia_stab(hwdesc->busid[hwdesc->bus_num - 1],
				    pcmcia_info, "network",
				    0, 4, 1)) {
		error_log(LOG_DEBUG,"Slot %s hosts interface %s\n",
			  hwdesc->busid[hwdesc->bus_num - 1], pcmcia_info);
		snprintf(interfacepath, SYSFS_PATH_MAX, "%s/%s/net/%s",
			 mnt_path, SYSFS_CLASS_NAME, pcmcia_info);
		snprintf(subdir, SYSFS_PATH_MAX, "%s/device", interfacepath);
		/* If there is _no_ link then it's a PC-Card and we accept
		 * what we found. */
		if (sysfs_get_link(subdir, devicelink, SYSFS_PATH_MAX) != 0) {
		    check_or_add_interface(hwdesc, interfacepath, iface_class_net);
		}
	    }
	}
	
	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
	 * 11) get buspaths from devicepath
	 * ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
	 * Currently we keep that mixed with 21)
	 * (whenever this is changed, remember the 'ccwgroup -> ccw' hack below)
	 */
	error_log(LOG_DEBUG,"Step 11: buspath from devicepath\n");

    } // if there is a devicepath

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
     * 21) get businfo from any path
     * ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
     * Now try to get bus information.
     * For that we need the devicepath.
     * 
     * We take the last part of the devicepath and search for it in
     * /sys/bus. Then check if the link found points to devicepath.
     *
     * Also look below class/pcmcia_socket/pcmcia_socket<N> if there is a
     * device link toi `dirname devicepath`. Then it is a pcmcia device in
     * slot N.
     */
    error_log(LOG_DEBUG,"Step 21: businfo from any path\n");

    lp = hwdesc->devicepath;
    sfc_class = sysfs_open_class("pcmcia_socket");
    if (sfc_class == NULL) {
	error_log(LOG_WARNING, "Could not open class %s\n",
		  "pcmcia_socket");
    } else {
	ifacelist = sysfs_get_class_devices(sfc_class);
	if (ifacelist == NULL) {
	    error_log(LOG_WARNING, "No interfaces in class %s\n",
		      "pcmcia_socket");
	}
    }
    hwdesc->bus_num = 0;

    while (lp != NULL && *lp != '\0') {
	rp = index(lp, '/');
	if (rp == lp) {
	    lp++;
	    continue;
	}
	if (rp != NULL) {
	    strncpy(devicename, lp, rp - lp);
	    devicename[rp-lp] = '\0';
	    strncpy(devicepath, hwdesc->devicepath, rp - hwdesc->devicepath);
	    devicepath[rp - hwdesc->devicepath] = '\0';
	    lp = rp + 1;
	} else {
	    strcpy(devicename, lp);
	    strcpy(devicepath, hwdesc->devicepath);
	    lp = NULL;
	}

	error_log(LOG_DEBUG,"businfo: %s (path %s)\n",
		  devicename, devicepath);

	if (!strcmp(devicename, "sys")) continue;
	if (!strcmp(devicename, SYSFS_DEVICES_NAME)) continue;
	
	/* Check for 'bus' link */
	snprintf(subdir, SYSFS_PATH_MAX, "%s/bus",
		 devicepath);
	if (sysfs_get_link(subdir, devicelink, SYSFS_PATH_MAX)) {
	    error_log(LOG_DEBUG,"\tno bus link\n", devicepath);
	    continue;
	}

	error_log(LOG_DEBUG,"buslink: %s\n", devicelink);
	busname = strrchr(devicelink,'/');
	if (busname)
	    busname ++;
	devname = devicename;

	error_log(LOG_DEBUG,"bus %s dev %s\n", busname, devname);

	/* If we have pcmcia sockets then search there as well for the device
	 * and if found, overwrite what we found while searching busses. */
  	if (devname && ifacelist != NULL) {
  	    dlist_for_each_data(ifacelist, sfcd_iface,
  				struct sysfs_class_device) {
		error_log(LOG_DEBUG,"interface: %s --- ", sfcd_iface->name);
		sfdv_device = sysfs_get_classdev_device(sfcd_iface);
  		if (sfdv_device == NULL) {
  		    error_log(LOG_INFO, "No device for iface %s %s\n",
  			      "pcmcia_socket", sfcd_iface->name);
  		} else {
		    error_log(LOG_DEBUG,"%s/%s --> %s\n", "pcmcia_socket",
			      sfcd_iface->name, sfdv_device->name);
  		    if (strcmp(devname, sfdv_device->name))
  			continue;
		    if ((hwdesc->bus_num > 0) &&
			!strcmp(busname,"pci") &&
			!strcmp(hwdesc->busname[hwdesc->bus_num - 1], "pci"))
			/* Disallow stacked PCI ids */
			hwdesc->bus_num--;
		    strncpy(hwdesc->busid[hwdesc->bus_num], devname,
			    MAX_HWDESC_PART_LEN);
		    strncpy(hwdesc->busname[hwdesc->bus_num], busname,
			    MAX_HWDESC_PART_LEN);
		    hwdesc->bus_num++;
		    devname = sfcd_iface->name
  			+ strcspn(sfcd_iface->name, "0123456789");
		    busname = (char *)hwdesc_bus_names[bus_pcmcia];
		    error_log(LOG_DEBUG,"Found %s device in slot %s\n", 
			      busname, devname);
  		}
  	    }
  	}

	/* If we found the bus for the device, add it to hwdesc. */
	if (devname) {
	    /* Adjust scsi busids:
	     * the number before the first colon is bogus */
	    if (!strcmp(busname, "scsi")) {
		/* Look for the first colon */
		s = index(devname, ':');
		if (s != NULL) {
		    /* Check if there are two more colons */
		    t = index(s+1, ':');
		    if (t != NULL && index(t+1, ':') != NULL)
			devname = s + 1;
		}
	    }
	    /* If busname is ccwgroup we always change that to ccw, because it
	     * still is ccw hardware. It is just grouped. */
	    if (!strcmp(busname, "ccwgroup"))
		busname = (char *)hwdesc_bus_names[bus_ccw];
	    if (hwdesc->bus_num < MAX_INTERFACE_NUM) {
		if ((hwdesc->bus_num > 0) &&
		    !strcmp(busname,"pci") &&
		    !strcmp(hwdesc->busname[hwdesc->bus_num - 1], "pci" )) {
		    /* Disallow stacked PCI ids */
		    hwdesc->bus_num--;
		} else {
		    strncpy(hwdesc->busname[hwdesc->bus_num], busname,
			    MAX_HWDESC_PART_LEN);
		}
		strncpy(hwdesc->busid[hwdesc->bus_num], devname,
			MAX_HWDESC_PART_LEN);
		hwdesc->bus_num++;
	    } else {
		error_log(LOG_ERR, "bus array not big enough\n");
	    }
	}
    }

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
     * 22) get id from any path
     * ^^^^^^^^^^^^^^^^^^^^^^^^
     * Look for address in interfacepath and for serial in devicepath and read
     * its content.
     */
    error_log(LOG_DEBUG,"Step 22: id from any path\n");

    char file[SYSFS_PATH_MAX] = {'\0'};
    if (*hwdesc->id == '\0') {
	/* If it is a class/net interface there might be a  attribute 'address'
	 * which contains the MAC address (if any). We have to check if it is
	 * not just zeroes. */
	i = 0;
	while (hwdesc->interfaceclass[i] > iface_class_net) i++;
	if ((hwdesc->interfaceclass[i] != iface_class_none) && (*hwdesc->interfacepath[i] != '\0')) {
	    error_log(LOG_DEBUG,"check for sysfs attribute 'address': ");
	    snprintf(subdir, SYSFS_PATH_MAX, "%s/address",
		     hwdesc->interfacepath[i]);
	    sfa_id = sysfs_open_attribute(subdir);
	    if (sfa_id != NULL) {
		error_log(LOG_DEBUG,"ok.\n");
		sysfs_read_attribute(sfa_id);
		if (sfa_id->value != NULL) {
//		    printf("address: %s %d\n", sfa_id->value,
//			   sfa_id->len);
		    for (i=0; i<sfa_id->len; i++) {
//			printf("-%c-\n", sfa_id->value[i]);
			switch (sfa_id->value[i]) {
			    case '\n':
				sfa_id->value[i] = '\0';
			    case ':':
			    case '0':
				continue;
			    default :
				is_id = 1;
			}
		    }
		}
		if (sfa_id->len >= MAX_HWDESC_PART_LEN)
		    error_log(LOG_ERR, "ID to long: %d\n", sfa_id->len);
		else
		    if (is_id)
			strcpy(hwdesc->id, sfa_id->value);
		sysfs_close_attribute(sfa_id);
	    } else {
		error_log(LOG_DEBUG,"not found.\n");
	    }
	}
    }
    if (*hwdesc->id == '\0') {
	/* Walk backwards through devicepath and look for an attribute
	 * 'serial'. */
	strcpy(subdir,hwdesc->devicepath);
	while (*subdir != '\0') {
	    /* Strip all trailing slashes. */
	    rp = subdir + strlen(subdir) - 1;
	    if (*rp == '/') {
		*rp = '\0';
		continue;
	    }
	    snprintf(file, SYSFS_PATH_MAX, "%s/serial", subdir);
	    sfa_id = sysfs_open_attribute(file);
	    if (sfa_id != NULL) {
		sysfs_read_attribute(sfa_id);
		if (sfa_id->value != NULL) {
//		    printf("device: %s\n", subdir);
//		    printf("serial: %s        %d\n", sfa_id->value,
//			   sfa_id->len);
		    if (sfa_id->len >= MAX_HWDESC_PART_LEN)
			error_log(LOG_ERR, "ID to long: %d\n", sfa_id->len);
		    else {
//			strcpy(hwdesc->id, sfa_id->value);
			/* Copy all ordinary characers; no newline */
			for (i=0; i<sfa_id->len; i++) {
			    if (sfa_id->value[i] < ' ')
				break;
			    hwdesc->id[i] = sfa_id->value[i];
			}
			hwdesc->id[i] = '\0';
		    }
		    sysfs_close_attribute(sfa_id);
		    break;
		}
		sysfs_close_attribute(sfa_id);
	    }
	    /* Now strip the last subdir */
	    rp = rindex(subdir, '/');
	    if (rp == NULL)
		break;
	    *rp = '\0';
	}
    }

    if ((*hwdesc->id == '\0') && (*hwdesc->devicepath != '\0')) {
	/*
	 * Second part of IUCV hack: 
	 * The ID of an IUCV device is always the link partner;
	 * this is stored in attribute 'user'.
	 */
	error_log(LOG_DEBUG,"Checking IUCV user (%s)\n", 
		  hwdesc->devicepath + 13);
	if (strlen(hwdesc->devicepath) > 17 &&
	    !strncmp(hwdesc->devicepath + 13,"iucv",4)) {
		strcpy(subdir,hwdesc->devicepath);
		snprintf(file, SYSFS_PATH_MAX, "%s/user", subdir);
		sfa_id = sysfs_open_attribute(file);
		if (sfa_id != NULL) {
		    sysfs_read_attribute(sfa_id);
		    if (sfa_id->value != NULL) {
			error_log(LOG_DEBUG,"device: %s\n", subdir);
			error_log(LOG_DEBUG,"user(%d): '%s'\n", sfa_id->len, 
				  sfa_id->value);
			if (sfa_id->len >= MAX_HWDESC_PART_LEN)
			    error_log(LOG_ERR, "ID to long: %d\n", sfa_id->len);
			else {
			    for (i=0; i<sfa_id->len; i++) {
				if (sfa_id->value[i] < ' ')
				    break;
				hwdesc->id[i] = sfa_id->value[i];
			    }
			    hwdesc->id[i] = '\0';
			}
		    }
		    sysfs_close_attribute(sfa_id);
		}
	}
    }
    error_log(LOG_DEBUG,"-> id %s\n", hwdesc->id);

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
     * 23) get vpid from devicepath
     * ^^^^^^^^^^^^^^^^^^^^^^^^^^^^
     */
    error_log(LOG_DEBUG,"Step 23: vpid from devicepath\n");

    if (*hwdesc->devicepath != '\0') {
	if (*hwdesc->vendorid == '\0') {
	    snprintf(file, SYSFS_PATH_MAX, "%s/vendor", hwdesc->devicepath);
//	    printf("vendorfile: %s\n", file);
	    sfa_id = sysfs_open_attribute(file);
	    if (sfa_id != NULL) {
		sysfs_read_attribute(sfa_id);
		if (sfa_id->value != NULL) {
//		    printf("device: %s\n", subdir);
//		    printf("serial: %s        %d\n", sfa_id->value,
//			   sfa_id->len);
		    if (sfa_id->len >= MAX_HWDESC_PART_LEN)
			error_log(LOG_ERR, "vendorid to long: %d\n",
				  sfa_id->len);
		    else {
//			strcpy(hwdesc->id, sfa_id->value);
			/* Copy all ordinary characers; no newline */
			for (i=0; i<sfa_id->len; i++) {
			    if (sfa_id->value[i] < ' ')
				break;
			    hwdesc->vendorid[i] = sfa_id->value[i];
			}
			hwdesc->vendorid[i] = '\0';
		    }
		}
		sysfs_close_attribute(sfa_id);
	    }

	}
	if (*hwdesc->vendorsubid == '\0') {
	    snprintf(file, SYSFS_PATH_MAX, "%s/subsystem_vendor",
		     hwdesc->devicepath);
	    sfa_id = sysfs_open_attribute(file);
	    if (sfa_id != NULL) {
		sysfs_read_attribute(sfa_id);
		if (sfa_id->value != NULL) {
		    if (sfa_id->len >= MAX_HWDESC_PART_LEN)
			error_log(LOG_ERR, "vendorsubid to long: %d\n",
				  sfa_id->len);
		    else {
			/* Copy all ordinary characers; no newline */
			for (i=0; i<sfa_id->len; i++) {
			    if (sfa_id->value[i] < ' ')
				break;
			    hwdesc->vendorsubid[i] = sfa_id->value[i];
			}
			hwdesc->vendorsubid[i] = '\0';
		    }
		}
		sysfs_close_attribute(sfa_id);
	    }

	}
	if (*hwdesc->productid == '\0') {
	    snprintf(file, SYSFS_PATH_MAX, "%s/device", hwdesc->devicepath);
	    sfa_id = sysfs_open_attribute(file);
	    if (sfa_id != NULL) {
		sysfs_read_attribute(sfa_id);
		if (sfa_id->value != NULL) {
		    if (sfa_id->len >= MAX_HWDESC_PART_LEN)
			error_log(LOG_ERR, "productid to long: %d\n",
				  sfa_id->len);
		    else {
			/* Copy all ordinary characers; no newline */
			for (i=0; i<sfa_id->len; i++) {
			    if (sfa_id->value[i] < ' ')
				break;
			    hwdesc->productid[i] = sfa_id->value[i];
			}
			hwdesc->productid[i] = '\0';
		    }
		}
		sysfs_close_attribute(sfa_id);
	    }

	}
	if (*hwdesc->productsubid == '\0') {
	    snprintf(file, SYSFS_PATH_MAX, "%s/subsystem_device",
		     hwdesc->devicepath);
	    sfa_id = sysfs_open_attribute(file);
	    if (sfa_id != NULL) {
		sysfs_read_attribute(sfa_id);
		if (sfa_id->value != NULL) {
		    if (sfa_id->len >= MAX_HWDESC_PART_LEN)
			error_log(LOG_ERR, "productsubid to long: %d\n",
				  sfa_id->len);
		    else {
			/* Copy all ordinary characers; no newline */
			for (i=0; i<sfa_id->len; i++) {
			    if (sfa_id->value[i] < ' ')
				break;
			    hwdesc->productsubid[i] = sfa_id->value[i];
			}
			hwdesc->productsubid[i] = '\0';
		    }
		}
		sysfs_close_attribute(sfa_id);
	    }

	}
    }

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
     * 24) get interface from interfacepath
     * ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
     */
    error_log(LOG_DEBUG,"Step 24: interface from interfacepath\n");

    ifnum = 0;
    while (hwdesc->interfaceclass[ifnum] != iface_class_none) {
	if (*hwdesc->interfacepath[ifnum] == '\0') { ifnum++; continue; }
	sysfs_get_name_from_path(hwdesc->interfacepath[ifnum],
				 hwdesc->interface[ifnum], SYSFS_PATH_MAX);
	strcpy(hwdesc->interfacetype[ifnum], 
		       get_interface_name(hwdesc->interfaceclass[ifnum]));
	ifnum++;
    }
    if (hwdesc->interface_num < ifnum) hwdesc->interface_num = ifnum;

#if 0
    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
     * 25) get type from any path
     * ^^^^^^^^^^^^^^^^^^^^^^^^^^
     * If we are in scope 'devices' then type will be the last busname.
     * if we are in scope 'interfaces' then type will be the interface class.
     * If don't know scope then look for an interface class and if we have no
     * interface take the last busname.
     * FIXME: set scope somehow, maybe as a second argument to this function
     * If interfaceclass is net then differ between different kind of network
     * interfaces:
     * - if we have net/<iface>/wireless --> type = wlan
     * - if ifacename starts with eth/tr --> type = eth/tr
     * - else --> type = net
     */
#ifdef DEBUG
    error_log(LOG_DEBUG,"Step 25: type from any path\n");
#endif
    scope_t scope = scope_none;
    if ( (*hwdesc->devtype == '\0') &&
	 (scope == scope_none || scope == scope_interface) &&
	 (*hwdesc->interfacepath != '\0') ) {
	sfcd_iface = sysfs_open_class_device_path(hwdesc->interfacepath);
	if (sfcd_iface != NULL) {
	    strncpy(hwdesc->devtype, sfcd_iface->classname,
		    MAX_HWDESC_PART_LEN);
	    sysfs_close_class_device(sfcd_iface);
	}
	if ( ! strcmp(hwdesc->devtype, "net")) {
	    s = strstr(hwdesc->interface[0], "eth");
	    if (s == hwdesc->interface[0])
		strcpy(hwdesc->devtype, "eth");
	    s = strstr(hwdesc->interface[0], "tr");
	    if (s == hwdesc->interface[0])
		strcpy(hwdesc->devtype, "tr");
	    snprintf(subdir, SYSFS_PATH_MAX, "%s/wireless",
		     hwdesc->interfacepath);
	    if ( ! sysfs_path_is_dir(subdir))
		strcpy(hwdesc->devtype, "wlan");
	}
    }
//    if ( (*hwdesc->devtype == '\0') &&
//	 (scope == scope_none || scope == scope_device) &&
//	 (hwdesc->bus_num > 0) ) {
//	strcpy(hwdesc->devtype, hwdesc->busname[hwdesc->bus_num-1]);
//    }

#endif
    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
     * 30) get driver name from devicepath
     *
     * Get the driver name, which might not be the module name
     */
    error_log(LOG_DEBUG,"Step 30: driver name from device path\n");

    if (*hwdesc->devicepath != '\0') {
	sfdv_device = sysfs_open_device_path(hwdesc->devicepath);
	if (sfdv_device != NULL) {
//	    printf("# devpath=%s\n# driver=%s\n",
//		   hwdesc->devicepath, sfdv_device->driver_name);
	    if (strcmp(sfdv_device->driver_name, "unknown"))
		strncpy(hwdesc->driver, sfdv_device->driver_name,
			MAX_DEVPATH_LEN);
	    sysfs_close_device(sfdv_device);
	}
    }


    return 0;
}
