/*
 * libgetconfig  (FIXME: NAME MAY STILL CHANGE)
 *
 * libgetconfig provides some functions to parse and assemble hardware
 * descriptions resp. configuration names. Some functions complete the
 * information we have so far. And one looks for a configuration that
 * matches best for the given hardware description.
 *
 * All the information resides in a struct hwdesc.
 *
 * It uses libsysfs to get info from sysfs.
 *
 * Version: $Id: get_config_lib.h 1433 2006-03-01 17:26:45Z zoz $
 */

#include <sysfs/libsysfs.h>

#ifndef _GET_CONFIG_LIB_H_
#define _GET_CONFIG_LIB_H_

#define MAX_HWDESC_LEN 256
#define MAX_HWDESC_PART_NUM 16
#define MAX_HWDESC_PART_LEN 64
#define MAX_INTERFACE_NUM   16
#define MAX_DEVPATH_LEN 256     /* !!!! replace with SYSFS_PATH_MAX !!!! */
#define MAX_PREFIX_LEN 16
/* We need to differ between the empty string and no string at all. Therefore
 * any string starting with 0 is no string and any string starting with
 * EMPTY_STRING is the empty string. 
 * This value must be smaller then any printable character. */
#define EMPTY_STRING 1

/* !!! The order of this enumeration is important!
 * When looking for available configurations for a hardware description we
 * might get different matches. These matches are rated by type and level.
 * The earlier a type is listed here the higher it will be rated.
 * The only exception is 'none', which must stay at first position and is
 * ignored when rating. The list must also end with 'invalid'.
 * And: (also important)
 * Keep the macro HWDESC_TYPE_NAMES in sync with hwdesc_type_t. It feeds a
 * static string array in get_config.c needed for detection of hwdesc_type
 * and also for proper output.
 */
typedef enum hwdesc_type_t {
    none,
    id,
    bus,
    devpath,
    interface,
    vpid,
    type,
    invalid
} hwdesc_type_t;
#define HWDESC_TYPE_NAMES \
    "not set", \
    "id", \
    "bus", \
    "devpath", \
    "interface", \
    "vpid", \
    "type", \
    "invalid"

typedef enum scope_t {
    scope_none,
    scope_device,
    scope_interface,
    scope_invalid
} scope_t;

typedef enum hwdesc_bus_t {
    bus_none,
    bus_pci,
    bus_scsi,
    bus_ide,
    bus_pcmcia,
    bus_usb,
    bus_firewire,
    bus_css,
    bus_ccw,
    bus_vio,
    bus_invalid
} hwdesc_bus_t;
#define HWDESC_BUS_NAMES \
    "none", \
    "pci", \
    "scsi", \
    "ide", \
    "pcmcia", \
    "usb", \
    "ieee1394", \
    "css", \
    "ccw", \
    "vio", \
    "invalid"

/* !!! If you add a new network class, do it _before_ 'net'! There are
 * !!! comparisons 'xxx <= iface_class_net'. */
typedef enum hwdesc_interface_class_t {
    iface_class_none,
    iface_class_eth,
    iface_class_tr,
    iface_class_wlan,
    iface_class_net,
    iface_class_block,
    iface_class_partition,
    iface_class_generic,
    iface_class_sound,
    iface_class_tty,
    iface_class_invalid
} hwdesc_interface_class_t;
#define HWDESC_INTERFACE_NAMES \
    "none", \
    "eth", \
    "tr", \
    "wlan", \
    "net", \
    "block", \
    "partition", \
    "generic", \
    "sound", \
    "tty", \
    "invalid"

static inline int match_interface_class(hwdesc_interface_class_t m1,
					hwdesc_interface_class_t m2)
{
    if (m1 != m2) {
	if ((m1 == iface_class_net ||
	     m1 == iface_class_eth ||
	     m1 == iface_class_tr ) &&
	    (m2 == iface_class_net ||
	     m2 == iface_class_eth ||
	     m2 == iface_class_tr)) {
	    return 1;
	} else {
	    return 0;
	}
    } else {
	return 1;
    }
}

/* This is the central data container. It contains
 * - the hardware description (that should be parsed or assembled
 * - all the descriptive hardware properties
 * - 3 devpath's (see below)
 * - the name of the configuration looked for in a certain directory
 */
typedef struct hwdesc_t {
    char hwdesc[MAX_HWDESC_LEN];
    char devtype[MAX_HWDESC_PART_LEN];
    char interface[MAX_INTERFACE_NUM][MAX_HWDESC_PART_LEN];
    int interface_num;
    char id[MAX_HWDESC_PART_LEN];
    char busname[MAX_INTERFACE_NUM][MAX_HWDESC_PART_LEN];
    char busid[MAX_INTERFACE_NUM][MAX_HWDESC_PART_LEN];
    int bus_num;
    char vendorid[MAX_HWDESC_PART_LEN];
    char productid[MAX_HWDESC_PART_LEN];
    char vendorsubid[MAX_HWDESC_PART_LEN];
    char productsubid[MAX_HWDESC_PART_LEN];
    char vpidrev[MAX_HWDESC_PART_LEN];
    char driver[MAX_HWDESC_PART_LEN];
    char config[MAX_INTERFACE_NUM][MAX_HWDESC_PART_LEN];
    int config_num;
    hwdesc_type_t hwdesc_type;
    /* devpath is what we may get from hotplug $DEVPATH. It may point to a real
     * device (/sys/devices/...) or to an interface (/sys/class/...,
     * /sys/block/...). Sometimes we need both. */
    char devpath[MAX_DEVPATH_LEN];
    /* devicepath is only a real devicepath (/sys/devices/...) if it is
     * available. It may be derived from the device link below the
     * interfacepath.*/
    char devicepath[MAX_DEVPATH_LEN];
    /* interfacecepath is the path below /sys/class or /sys/block to the
     * interface of the device. */
    char interfacepath[MAX_INTERFACE_NUM][MAX_DEVPATH_LEN];
    /* interfaceclass is the type of the interface */
    hwdesc_interface_class_t interfaceclass[MAX_INTERFACE_NUM];
    /* interfacetype is the name of the interface type */
    char interfacetype[MAX_INTERFACE_NUM][MAX_DEVPATH_LEN];
} hwdesc_t;

/* It splits a hardware description into it's parts. To use it properly you
 * should set the whole struct hwdesc to zero and put then your hardware
 * description to hwdesc->hwdesc.
 * When it has finished the struct has been filled with the information from
 * the given hwdesc.
 * Return values are still under construction. Currently (as usual)
 * 0: success
 * !=0: failure
 * -101: to much characters in a part of a hardware description
 * -102: to much parts in a hardware description
 * +111: invalid hardware description
 * If we have to differ between differnt failures it may be enhanced
 */
int split_hwdesc(hwdesc_t* hwdesc);

/* assemble_hwdesc() creates a new hardware description out of the information
 * inside the given struct hwdesc. In case of success it writes the new
 * description as well into hwdesc->hwdesc. The type of the new description is
 * also taken from the struct (hwdesc->hwdesc_type).
 * struct->hwdesc will not be written if some error occurs.
 * Except for types 'interface' and 'type', the 'devtype' will be added as
 * first part, if it is available. No error, if it is missing.
 * It returns:
 * 0 if success
 * -1 if there was no valid type for the new description
 * >0 if some needed data was missing
 * -512 if the new description would exceed MAX_HWDESC_LEN characters
 * -511, -510, ... like -512 and some data missing
 */
int assemble_hwdesc(hwdesc_t* hwdesc);

/* It looks for a configuration file starting with 'prefix' in 'directory'
 * that matches best the hardware description 'hwdesc'.
 * In case of success the name of the configuration (not the file, i.e.
 * without the prefix) will be written into hwdesc->config[0].
 * Return values:
 * 0 if success
 * <0 if internal error
 * >0 no config found
 */
int get_configuration(hwdesc_t* hwdesc,
		      const char* directory, const char* prefix);

/* Here we try to get some more information from sysfs. We try to get
 * hwdesc->devicepath and hwdesc->interfacepath.
 * Then we can get much more information out of sysfs.
 *
 * This function is not yet finished: currently it only gets the interfacepath
 * and the devicepath. Will be completed very soon.
 */
int complete_hwdesc_sysfs(hwdesc_t* hwdesc);

/*
 * Matches a given interface name to a class.
 * Returns hwdesc_class_none if iface is zero,
 * hwdesc_class_invalid if no match was found,
 * the found hwdesc_class_t otherwise.
 */
hwdesc_interface_class_t match_interface_type(const char *iface);
char *get_interface_name( hwdesc_interface_class_t class );
void print_hwdesc(hwdesc_t *hwdesc, char *prefix, int export);

/* temporary */
int match_type(const char *cfg, const char *hwd);
int get_all_interfaces(const char *classname, const char *configname,
		       char *interfaces);
int look_smpppd_info(char *name);


/* There might be some more functions as for example
 * - get_interface_name() that gets the currently used interface name for a
 *   hwdesc. Usefull if you refer to an 'interface' pci-2-3 inside a
 *   configuration (the recommended way)
 * - get_device() the same vice versa
 * - get_options() get the options that were used by ifup/hwup for a given
 *   hwdesc. Don't know if we need that; perhaps Arvin.
 *
 * If somebody needs more, tell zoz@suse.de
 */

#endif //ifndef _GET_CONFIG_LIB_H_
